'''
When PyThea is installed in an enviroment,
with this script you can directry call the package as an excecutable.
$ python3 -m pip install ./ --use-feature=in-tree-build
$ pythea streamlit
'''
import os

import click
import streamlit.cli

from .version import version


@click.group()
def main():
    pass


@main.command('streamlit')
def main_streamlit():
    """Run the main program in browser."""
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'PyThea_app.py')
    args = []
    streamlit.cli._main_run(filename, args)


@main.command('help')
@click.pass_context
def help(ctx):
    """Print this help message."""
    # For 'pythea --help' instead of 'pythea help'.
    import sys
    sys.argv[1] = '--help'
    main()


@main.command('version')
def version():
    """Print PyThea's version number."""
    print(f'PyThea installed version is: {version}')


@main.command('docs')
def docs():
    """Show documents in browser."""
    print('Showing document page in browser is not implemented yet.')


@main.command('update')
def update():
    """Update PyThea."""
    print('Not implemented yet.')

# TODO Include the test here.


if __name__ == '__main__':
    main()



'''
Configuration for the seleced bodies.
'''

bodies_dict = {'Mercury': ['199', 'darkturquoise'],
               'Venus': ['299', 'darkorchid'],
               'Earth': ['399', 'green'],
               'Mars': ['499', 'maroon'],
               'Jupiter': ['599', 'navy'],
               'STEREO B': ['-235', 'b'],
               'STEREO A': ['-234', 'red'],
               'Solar Orbiter': ['-144', 'dodgerblue'],
               'Parker Solar Probe': ['-96', 'purple'],
               'BepiColombo': ['-121', 'orange']}



import pkgutil


def test_import_main():
    """
    This imports PyThea
    """
    no_requests = False
    try:
        pass
    except ImportError:
        no_requests = True
    assert no_requests is False


def test_imports_all():
    """
    This imports all modules in PyThea
    """
    def on_error():
        try:
            raise
        except Warning:
            pass

    for imper, nm, ispkg in pkgutil.walk_packages(['PyThea'], 'PyThea.',
                                                  onerror=on_error):
        imper.find_spec(nm)



"""
Test the utilities
"""

from sunpy.net import hek

from PyThea.utils import get_hek_flare


# ------------------
# Test utilities
def test_get_hek_flare_HEKTable():
    """
    Test that the get_hek_flare returns a hek.hek.HEKTable
    """
    from datetime import datetime
    day = datetime(2017, 9, 10)
    _, flare_list_ = get_hek_flare(day)

    assert type(flare_list_) == hek.hek.HEKTable



"""
    PyThea: A software package to reconstruct the 3D structure of CMEs and
    shock waves using multi-viewpoint remote-sensing observations.
    Copyright (C) 2021  Athanasios Kouloumvakos

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import astropy.units as u
import numpy as np
import streamlit as st
from astropy.coordinates import Distance, SkyCoord, SphericalRepresentation
from callbacks import load_or_delete_fittings
from config import (app_styles, config_sliders, selected_bodies,
                    selected_imagers)
from extensions.stqdm import stqdm  # See also https://github.com/tqdm/tqdm
from geometrical_models import ellipsoid, gcs, spheroid
from modules import (date_and_event_selection, final_parameters_gmodel,
                     fitting_and_slider_options_container, fitting_sliders,
                     maps_clims)
from sunpy.coordinates import frames
from sunpy_dev.map.maputils import get_closest
from utils import (download_fits, make_figure, maps_process, model_fittings,
                   plot_bodies, plot_fitting_model)
from version import version


def delete_from_state(vars):
    for var in vars:
        if var in st.session_state:
            del st.session_state[var]


def footer_text():
    st.subheader('About this application:')
    st.markdown("""
                   _PyThea_  is an open-source software package that can be used to
                   reconstruct the 3D structure of Coronal Mass Ejections (CMEs) and
                   shock waves and determine their kinematics using remote-sensing observations.
                   The tool implements the Graduate Cylindrical Shell (GCS) model that can be used
                   to reconstruct CMEs and two geometrical models, namely a spheroid and ellipsoid model
                   to reconstruct shock waves. It also implements remote-sensing observations
                   from multiple viewpoints such as the SOlar and Heliospheric Observatory (SOHO)
                   and Solar Terrestrial Relations Observatory (STEREO).
                """)
    right, left = st.columns((1, 1))
    right.markdown("""
                   **Github**: Find the latest version here
                               [![GitHub](https://img.shields.io/badge/github-%23121011.svg?style=for-the-badge&logo=github&logoColor=white)](https://github.com/AthKouloumvakos/PyThea) \n
                   **Citation**: Cite PyThea as [![https://doi.org/10.5281/zenodo.5713659](https://zenodo.org/badge/DOI/10.5281/zenodo.5713659.svg)](https://doi.org/10.5281/zenodo.5713659) \n
                   """ +
                   f"""
                   **Version**: {version} (latest release [![Version](https://img.shields.io/github/v/release/AthKouloumvakos/PyThea)](https://github.com/AthKouloumvakos/PyThea/releases))
                   """)
    left.image('https://github.com/AthKouloumvakos/PyThea/blob/master/docs/logo/pythea_logo.png?raw=true')
    st.markdown('---')


def run():
    ###################################################################
    # IMPORTANT: DON'T CHANGE THE POSSITION OF THE COMPONENTS. NEVER! #
    ###################################################################

    #############################################################
    # set page config
    st.set_page_config(page_title='PyThea', page_icon=':rocket:',
                       initial_sidebar_state='expanded')

    #############################################################
    # HTML Styles
    app_styles.apply(st)

    #############################################################
    # Main page information text
    st.title('PyThea: Reconstruct CMEs & Shocks')

    #############################################################
    # Startup Variables
    if 'startup' not in st.session_state:
        st.session_state.startup = {'fitting': True}

    #############################################################
    # Date and Event selection

    if 'date_process' not in st.session_state:
        date_and_event_selection(st)
        st.markdown(""" #### ⏻ Select a day & solar event. """)
    else:
        st.sidebar.markdown('## Processing Event|Date:')
        st.sidebar.info(f'{st.session_state.event_selected}')

    if 'date_process' not in st.session_state:
        st.markdown('---')
        footer_text()
        st.stop()

    #############################################################
    # 3D Model Selection
    st.sidebar.markdown('## 3D Fitting and Reconstruction')

    if 'geometrical_model' not in st.session_state:
        st.markdown(""" #### ⏻ Select the geometrical model you want to fit. """)
        geometrical_model = st.sidebar.selectbox('Geometrical model to fit',
                                                 options=['Select a model', 'Spheroid', 'Ellipsoid', 'GCS'])
        if geometrical_model != 'Select a model':
            st.session_state.geometrical_model = geometrical_model
            st.experimental_rerun()
    else:
        st.sidebar.info(f'Geometrical Model: \
                        {st.session_state.geometrical_model}')

    if 'geometrical_model' not in st.session_state:
        st.markdown('---')
        footer_text()
        st.stop()

    #############################################################
    # 3D Fitting and Reconstruction
    st.markdown('---')
    fitting_and_slider_options_container(st)

    if st.session_state.coord_system == 'HGC':
        long_val = [0., 360.]
    elif st.session_state.coord_system == 'HGS':
        long_val = [-180., 180.]

    if 'longit' not in st.session_state:
        st.session_state.longit = 0
    longit = st.sidebar.slider(f'{st.session_state.coord_system} \
                               Longitude [deg.]:',
                               min_value=long_val[0],
                               max_value=long_val[1],
                               step=0.01, key='longit') * u.degree

    if 'latitu' not in st.session_state:
        st.session_state.latitu = 0
    latitu = st.sidebar.slider(f'{st.session_state.coord_system} \
                               Latitude [deg.]:',
                               min_value=-90.,
                               max_value=90.,
                               step=0.01, key='latitu') * u.degree

    fitting_sliders(st)

    col1, col3, col2 = st.sidebar.columns(3)
    store_fit_button_pressed = st.sidebar.button('Store Fit')

    #############################################################
    # 3D Fitting and Reconstruction
    st.sidebar.markdown('## Imaging menu')
    with st.sidebar.expander('Download Options'):
        select_imagers_form = st.form(key='select_imagers_form')
        imagers_list = select_imagers_form.multiselect('Select Imagers',
                                                       options=selected_imagers.imager_dict.keys(),
                                                       default=['LC2', 'LC3', 'COR2A'],
                                                       key='imagers_list')
        select_imagers_form.form_submit_button(label='Submit',
                                               on_click=delete_from_state,
                                               kwargs={'vars': ['map', ]})
        select_timerange_form = st.form(key='select_timerange_form')
        imaging_time_range = select_timerange_form.slider('Time Range [hours]',
                                                          -1., 6., [-1., 1.], 0.5,
                                                          key='imaging_time_range')
        select_timerange_form.form_submit_button(label='Submit',
                                                 on_click=delete_from_state,
                                                 kwargs={'vars': ['map', 'map_']})

    with st.sidebar.expander('Processing Options'):
        procoption_container = st.container()
        if 'imagers_list_' not in st.session_state:
            # imagers_list_ is used later when we download or filter the images
            st.session_state['imagers_list_'] = []
        image_mode = procoption_container.selectbox('Map sequence processing',
                                                    options=['Running Diff.', 'Base Diff.', 'Plain'], key='image_mode',
                                                    on_change=delete_from_state, kwargs={'vars': ['map', ]})

    with st.sidebar.expander('Plot/View Options'):
        plotviewopt_container = st.container()
        clip_model = plotviewopt_container.checkbox('Clip plot on image limits', value=True)
        plt_supp_imagers = plotviewopt_container.checkbox('Supplementary Imaging', value=False)
        star_field = plotviewopt_container.checkbox('View Bodies or s/c')
        if star_field is True:
            bodies_list = plotviewopt_container.multiselect('Select Bodies', options=selected_bodies.bodies_dict.keys(),
                                                            default=['Mercury', 'Venus', 'Jupiter'])

    #############################################################
    # Download and Process the Images
    # This part runs only if the map_ doesn't exits or the session_state.map_ does not contain all the imagers requested
    if 'map_' not in st.session_state or [False for lst in imagers_list if lst not in st.session_state.map_]:
        st.session_state.map_ = {} if 'map_' not in st.session_state else st.session_state.map_
        progress_bar = stqdm.stqdm(imagers_list, desc='Preparing to Download data')
        for imager in progress_bar:
            if imager in st.session_state.map_:
                pass
            else:
                progress_bar.desc = f'Downloaded {imager} images from VSO'
                st.session_state.map_[imager] = download_fits(st.session_state.date_process,
                                                              imager, time_range=imaging_time_range)

    if 'map' not in st.session_state:
        st.session_state.map, st.session_state.imagers_list_ = maps_process(st.session_state.map_,
                                                                            imagers_list,
                                                                            image_mode)
        maps_clims(st, imagers_list)

    if not st.session_state.imagers_list_:
        st.error('No images have been downloaded or processed.')  # TODO: Explain better
        st.stop()

    #############################################################
    # Selection for the primary image to plot
    st.markdown('### Multi-viewpoint imaging')
    col1, col2 = st.columns((1, 3))
    imager_select = col1.selectbox('Select an imager',
                                   options=st.session_state.imagers_list_)

    maps_date = [maps.date for maps in st.session_state.map[imager_select]]
    if len(maps_date) > 1:
        running_map_date = col2.select_slider('Slide to the image time',
                                              options=maps_date, key='running_map_date')
    else:
        running_map_date = maps_date[0]

    running_map = st.session_state.map[imager_select][maps_date.index(running_map_date)]

    qmin = st.session_state.map_clim[imager_select][0]  # np.nanquantile(running_map.data, 0.20)
    qmax = st.session_state.map_clim[imager_select][1]  # np.nanquantile(running_map.data, 0.80)
    col1, col2 = st.columns((1, 3))
    cmmin, cpmax = config_sliders.slider_image_pmclims[image_mode]
    clim = plotviewopt_container.slider('Images climits:',
                                        float(cmmin), float(cpmax),
                                        (float(qmin-10), float(qmax+10)),
                                        key='clim')

    #############################################################
    # Finalize the geometrical model

    if st.session_state.geometrical_model == 'Spheroid':
        rcenter, radaxis, orthoaxis1 = final_parameters_gmodel(st)
    elif st.session_state.geometrical_model == 'Ellipsoid':
        rcenter, radaxis, orthoaxis1, orthoaxis2, tilt = final_parameters_gmodel(st)
    elif st.session_state.geometrical_model == 'GCS':
        rcenter, height, alpha, kappa, tilt = final_parameters_gmodel(st)

    Spher_rep = SphericalRepresentation(longit, latitu,
                                        Distance(np.abs(rcenter)))
    if st.session_state.coord_system == 'HGC':
        center = SkyCoord(np.sign(rcenter) * Spher_rep.to_cartesian(),
                          frame=frames.HeliographicCarrington,
                          observer='earth',
                          obstime=running_map.date)
    elif st.session_state.coord_system == 'HGS':
        center = SkyCoord(np.sign(rcenter) * Spher_rep.to_cartesian(),
                          frame=frames.HeliographicStonyhurst,
                          observer='earth',
                          obstime=running_map.date)
    st.session_state.center = center

    if st.session_state.geometrical_model == 'Spheroid':
        model = spheroid(center, radaxis, orthoaxis1)
    elif st.session_state.geometrical_model == 'Ellipsoid':
        model = ellipsoid(center, radaxis, orthoaxis1, orthoaxis2, tilt)
    elif st.session_state.geometrical_model == 'GCS':
        model = gcs(center, height, alpha, kappa, tilt)

    #############################################################
    # Plot main and supplement figure images
    fig, axis = make_figure(running_map, image_mode, clim=clim, clip_model=clip_model)
    if st.session_state.plot_mesh_mode == 'Skeleton':
        model.plot(axis, mode='Skeleton')
    if st.session_state.plot_mesh_mode == 'Full':
        model.plot(axis, mode='Skeleton')
        model.plot(axis, mode='Full')
    if st.session_state.plot_mesh_mode == 'Surface':
        model.plot(axis, only_surface=True)

    if star_field is True:
        plot_bodies(axis, bodies_list, running_map)
        axis.legend()
    st.pyplot(fig)

    if plt_supp_imagers and len(st.session_state.imagers_list_) > 2:
        supl_imagers = st.select_slider('Select supplement imagers',
                                        options=st.session_state.imagers_list_,
                                        value=(st.session_state.imagers_list_[1],
                                               st.session_state.imagers_list_[-1]),
                                        key='supl_imagers')
        col1, col2 = st.columns(2)
        fig, axis = make_figure(get_closest(st.session_state.map[supl_imagers[0]],
                                running_map_date), image_mode)
        model.plot(axis, mode='Skeleton')
        col1.pyplot(fig)
        fig, axis = make_figure(get_closest(st.session_state.map[supl_imagers[1]],
                                            running_map_date), image_mode)
        model.plot(axis, mode='Skeleton')
        col2.pyplot(fig)

    #############################################################
    # Store the fitting
    single_fit = model.to_dataframe()
    single_fit['imager'] = imager_select

    if store_fit_button_pressed:
        if 'model_fittings' not in st.session_state:
            st.session_state.model_fittings = model_fittings(st.session_state.event_selected,
                                                             st.session_state.date_process.strftime('%Y-%m-%dT%H:%M:%S.%f'),
                                                             st.session_state.geometrical_model,
                                                             single_fit)
        else:
            st.session_state.model_fittings.parameters = single_fit.combine_first(st.session_state.model_fittings.parameters)

    #############################################################
    # View the fittings table
    if st.session_state.fitting_table is True:
        st.markdown('---')
        st.markdown('### Parameters Table')
        st.markdown('**Running Fitting Table:**')
        st.dataframe(single_fit)
        if 'model_fittings' in st.session_state:
            st.markdown('**Stored Fitting Table:**')
            st.dataframe(st.session_state.model_fittings.parameters)
            col1, col2 = st.columns((2, 2))
            col2.selectbox('Select a fitting',
                           options=st.session_state.model_fittings.parameters.index,
                           key='fitting_select')
            if 'fit_action' not in st.session_state:
                st.session_state.fit_action = 'Select'
            col1.selectbox('Action',
                           options=['Select', 'Load', 'Delete'],
                           on_change=load_or_delete_fittings, args=[st],
                           key='fit_action')

    #############################################################
    # Plot the kinematics
    if ('model_fittings' in st.session_state) and \
       (st.session_state.plt_kinematics is True):
        st.markdown('---')
        st.markdown('### Plots of kinematics ')
        col1, col2, col3 = st.columns(3)
        plt_kinematics_select = col1.selectbox('Select Plots',
                                               options=['All', 'HeightT', 'SpeedT', 'Long-LatT'])

        if 'fit_mode' not in st.session_state:
            st.session_state.fit_mode = 'polynomial'
        fit_mode = col2.selectbox('Select Fitting Mode',
                                  options=['polynomial', 'spline', 'custom'],
                                  key='fit_mode')
        if fit_mode == 'polynomial':
            if 'polyfit_order' not in st.session_state:
                st.session_state.polyfit_order = 2
            polyfit_order = col3.slider('Polynomial order', min_value=1, max_value=4, step=1, key='polyfit_order')
            fit_args_ = {'type': 'polynomial', 'order': polyfit_order}
        elif fit_mode == 'spline':
            if 'splinefit_order' not in st.session_state:
                st.session_state.splinefit_order = 3
            splinefit_order = col3.slider('Spline order', min_value=1, max_value=5, step=1, key='splinefit_order')
            if 'splinefit_smooth' not in st.session_state:
                st.session_state.splinefit_smooth = 0.5
            splinefit_smooth = st.slider('Spline smooth', min_value=0., max_value=1., step=0.01, key='splinefit_smooth')
            fit_args_ = {'type': 'spline', 'order': splinefit_order, 'smooth': splinefit_smooth}
        elif fit_mode == 'custom':
            if 'splinefit_smooth' not in st.session_state:
                st.session_state.fitcustexpres_select = 'a*exp(-b*x)+c'
            fitcustexpres_select = col3.selectbox('Select a custom function',
                                                  options=['a*exp(-b*x)+c'],  # 'a*sqrt(b*x)+c',  '((x+a)/(x+b))+c'
                                                  key='plt_fitcustexpres_select')
            fit_args_ = {'type': 'custom', 'expression': fitcustexpres_select,
                         'bounds': ([-np.inf, -np.inf, -np.inf], [np.inf, np.inf, np.inf]), 'order': 3}
            st.info(f'A custom function {fit_args_["expression"]} was fitted when processing parameters.')

        if plt_kinematics_select == 'All':
            col1, col2 = st.columns(2)
            fig_ht = plot_fitting_model(st.session_state.model_fittings,
                                        fit_args=fit_args_,
                                        plt_type='HeightT')
            col1.pyplot(fig_ht)
            fig_vt = plot_fitting_model(st.session_state.model_fittings,
                                        fit_args=fit_args_,
                                        plt_type='SpeedT')
            col2.pyplot(fig_vt)
        elif plt_kinematics_select == 'Long-LatT':
            fit_args_['bounds'] = ([-np.inf, 0, -np.inf], [np.inf, np.inf, np.inf])
            col1, col2 = st.columns(2)
            fig_loT = plot_fitting_model(st.session_state.model_fittings,
                                         fit_args=fit_args_,
                                         plt_type='LongT')
            col1.pyplot(fig_loT)
            fig_laT = plot_fitting_model(st.session_state.model_fittings,
                                         fit_args=fit_args_,
                                         plt_type='LatT')
            col2.pyplot(fig_laT)
        else:
            fig = plot_fitting_model(st.session_state.model_fittings,
                                     fit_args=fit_args_,
                                     plt_type=plt_kinematics_select)
            st.pyplot(fig)
        st.session_state.model_fittings.kinematics['fit_method'] = fit_args_
    else:
        st.session_state.startup['fitting'] = True

    #############################################################
    # Download Fitting and Figures
    st.sidebar.markdown('## Finalize and save results')
    if 'model_fittings' in st.session_state:
        st.sidebar.download_button('Download Fitting as .json file',
                                   st.session_state.model_fittings.to_json(),
                                   st.session_state.model_fittings.model_id()+'.json')
    else:
        st.sidebar.info('Store a fit to enable this feature.')
    st.markdown('---')
    # footer_text()


if __name__ == '__main__':
    run()



"""
    PyThea: A software package to reconstruct the 3D structure of CMEs and
    shock waves using multi-viewpoint remote-sensing observations.
    Copyright (C) 2021  Athanasios Kouloumvakos

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import datetime
import io
import json
from copy import copy

import astropy.units as u
import matplotlib.colors as colors
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import sunpy.map
from config.selected_bodies import bodies_dict
from config.selected_imagers import imager_dict
from scipy.interpolate import UnivariateSpline
from sunpy.coordinates import get_horizons_coord
from sunpy.map.maputils import contains_coordinate
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy_dev.map.maputils import (filter_maps, maps_sequence_processing,
                                    prepare_maps)
from version import version


def get_hek_flare(day):
    '''
    Returns from HEK the flare list for a given day
    '''
    flare_list = Fido.search(a.Time(day, day + datetime.timedelta(days=1)),
                             a.hek.EventType('FL'),
                             a.hek.FL.GOESCls > 'B1.0',
                             a.hek.OBS.Observatory == 'GOES')
    if len(flare_list['hek']) == 0:
        selectbox_list = ['No events returned', ]
        flare_list_ = []
    else:
        flare_list_ = flare_list['hek']['event_starttime', 'event_peaktime',
                                        'event_endtime', 'fl_goescls',
                                        'fl_peakflux', 'hgs_x', 'hgs_y', 'ar_noaanum']
        selectbox_list = []
        for flares in flare_list_:
            fl_ = flares['fl_goescls']
            t_ = flares['event_peaktime'].strftime('%Y-%m-%dT%H:%M:%S')
            selectbox_list.append((f'FL{fl_}|{t_}'))

    return selectbox_list, flare_list_


def make_figure(map, image_mode, clim=[-20, 20], clip_model=True):
    '''
    Makes the main imager figure and returns the figure and axis handle.
    '''
    fig = plt.figure()
    axis = plt.subplot(projection=map)
    # TODO: For plain images or when EUVIA-B are used, this does not work very well.
    if image_mode == 'Plain':
        map.plot()
    else:
        map.plot(cmap='Greys_r',
                 norm=colors.Normalize(vmin=clim[0], vmax=clim[1]))
    map.draw_limb()
    # map.draw_grid(linewidth=2, color='red') # TODO: This takes too much computation time. Maybe for AIA or EUVI?
    yax = axis.coords[1]
    yax.set_ticklabel(rotation=90)
    if clip_model:
        axis.set_xlim([0, map.data.shape[0]])
        axis.set_ylim([0, map.data.shape[1]])

    return fig, axis


def plot_bodies(axis, bodies_list, smap):
    '''
    Plots in the images the possition of the pre-configured bodies (Earth, STA, Venus etc.)
    '''
    for body in bodies_list:
        body_coo = get_horizons_coord(bodies_dict[body][0], smap.date)
        if contains_coordinate(smap, body_coo):
            axis.plot_coord(body_coo, 'o', color=bodies_dict[body][1],
                            fillstyle='none', markersize=6, label=body)


def download_fits(date_process, imager, time_range=[-1, 1]):
    '''
    Downloads the imaging data (fits files) from VSO
    '''
    timerange = a.Time(date_process + datetime.timedelta(hours=time_range[0]),
                       date_process + datetime.timedelta(hours=time_range[1]))

    map_ = {}
    args = imager_dict[imager][0]
    result = Fido.search(timerange, *args)
    print(result)
    if len(result['vso']) != 0:
        downloaded_files = Fido.fetch(result)
        try:
            map_ = sunpy.map.Map(downloaded_files)
        except RuntimeError as err:
            print('Handling RuntimeError error:', err)
            map_ = []
        except OSError as err:
            print('Handling OSError error:', err)
            map_ = []
    else:
        map_ = []

    return map_


def maps_process(ninstr_map_in, imagers_list_in, image_mode):
    '''
    Process the images for the selected imagers and return the final maps.

    Note
    ----
    Here the ninstr_map_in is the session_state.map_ when used from the application.
    '''
    ninstr_map_out = {}
    imagers_list_out = []

    for imager in imagers_list_in:
        extra = imager_dict[imager][1]
        if imager in ninstr_map_in and ninstr_map_in[imager] != []:
            ninstr_map_out[imager] = filter_maps(ninstr_map_in[imager], extra)
            ninstr_map_out[imager] = prepare_maps(ninstr_map_out[imager], extra)
            ninstr_map_out[imager] = maps_sequence_processing(ninstr_map_out[imager],
                                                              seq_type=image_mode)
            if ninstr_map_out[imager] != []:
                imagers_list_out.append(imager)

    return ninstr_map_out, imagers_list_out


# TODO: Implement units here
class model_fittings:
    '''
    A class to store the fittings of the geometrical model.
    '''

    def __init__(self, event_selected, date_process, geometrical_model, model_parameters, kinematics={'fit_method': None}):
        self.event_selected = event_selected
        self.date_process = date_process
        self.geometrical_model = geometrical_model
        self.parameters = model_parameters
        self.kinematics = kinematics

    def model_id(self):
        str_id = self.event_selected.replace('-', '').replace(':', '').replace('|', 'D').replace('.', 'p') + 'M' + self.geometrical_model
        return str_id

    def to_dict(self):
        parameters = copy(self.parameters)
        parameters['time'] = parameters.index.strftime('%Y-%m-%dT%H:%M:%S.%f')
        parameters = parameters.to_dict(orient='list')
        model_fittings_dict = {'event_selected': self.event_selected,
                               'date_process': self.date_process,
                               'geometrical_model': {'type': self.geometrical_model,
                                                     'parameters': parameters},
                               'kinematics': self.kinematics
                               }
        return model_fittings_dict

    def to_json(self, buffer=False):
        '''
        Returns the fittings of the geometrical model as a json format
        '''
        model_dict = self.to_dict()
        model_dict.update({'date_created': (datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:%S')})
        model_dict.update({'version': version})
        if buffer:
            json_buffer = io.BytesIO()
            return json_buffer.write(json.dumps(model_dict, indent=' ').encode())
        else:
            return json.dumps(model_dict, indent=' ')


def plot_fitting_model(model, fit_args, plt_type='HeightT'):
    '''
    Plot the height(speed)--time evolution of the fitting parameters.
    '''
    palete = sns.color_palette('deep')
    parameters = {
        'Spheroid':
            {'height': ['+', '', palete[3], 'h-apex'],
             'orthoaxis1': ['x', '', palete[0], 'r-axis1'],
             },
        'Ellipsoid':
            {'height': ['+', '', palete[3], 'h-apex'],
             'orthoaxis1': ['x', '', palete[0], 'r-axis1'],
             'orthoaxis2': ['x', '', palete[2], 'r-axis2'],
             },
        'GCS':
            {'height': ['+', '', palete[3], 'h-apex'],
             'rapex': ['x', '', palete[0], 'r-apex'],
             },
    }
    parameters = parameters[model.geometrical_model]
    # Height vs time
    fig = plt.figure(figsize=(5.5, 5.5), tight_layout=True)
    axis = plt.subplot()
    if plt_type == 'HeightT':
        for p in parameters.keys():
            plt.plot(model.parameters.index,
                     model.parameters[p],
                     marker=parameters[p][0],
                     linestyle=parameters[p][1],
                     color=parameters[p][2],
                     label=parameters[p][3])  # label='fit: a=%5.3f, b=%5.3f, c=%5.3f' % tuple(popt)
            if len(model.parameters[p])-1 > fit_args['order']:
                # How to get confidence intervals from curve_fit?
                fit = parameter_fit(model.parameters.index, model.parameters[p], fit_args)
                plt.plot(fit['best_fit_x'], fit['best_fit_y'], '-', color=parameters[p][2])
                plt.fill_between(fit['best_fit_x'], fit['sigma_bounds']['up'], fit['sigma_bounds']['low'],
                                 color=parameters[p][2], alpha=0.20)
                if fit_args['type'] == 'spline':
                    plt.fill_between(fit['best_fit_x'], fit['sigv_bounds']['up'], fit['sigv_bounds']['low'],
                                     color=parameters[p][2], alpha=0.05)
            else:
                plt.plot(model.parameters.index, model.parameters[p], '--', color=parameters[p][2])
        ylabel = 'Height or Length [Rsun]'
        plt.gca().set_ylim(bottom=0)
    elif plt_type == 'SpeedT':
        for p in parameters.keys():
            if len(model.parameters[p])-1 > fit_args['order']:
                # How to get confidence intervals from curve_fit?
                fit = parameter_fit(model.parameters.index, model.parameters[p], fit_args)
                Rs2km = (1 * u.R_sun).to_value(u.km)
                sec = 24*60*60
                speed_best_fit = (Rs2km/sec) * np.gradient(fit['best_fit_y'], fit['best_fit_x_num'])
                speed_bound_upper = (Rs2km/sec) * np.gradient(fit['sigma_bounds']['up'], fit['best_fit_x_num'])
                speed_bound_lower = (Rs2km/sec) * np.gradient(fit['sigma_bounds']['low'], fit['best_fit_x_num'])
                plt.plot(fit['best_fit_x'], speed_best_fit, '-', color=parameters[p][2], label=parameters[p][3])
                plt.fill_between(fit['best_fit_x'], speed_bound_lower, speed_bound_upper,
                                 color=parameters[p][2], alpha=0.20)
                if fit_args['type'] == 'spline':
                    plt.fill_between(fit['best_fit_x'], (Rs2km/sec) * fit['sigv_bounds']['dlow'], (Rs2km/sec) * fit['sigv_bounds']['dup'],
                                     color=parameters[p][2], alpha=0.05)
            else:
                pass
        ylabel = 'Speed [km/s]'
        plt.gca().set_ylim(bottom=0)
    if plt_type == 'LongT' or plt_type == 'LatT':
        parameters = {'LongT': ['hgln', '+', palete[3], 'Longitude'],
                      'LatT': ['hglt', '+', palete[2], 'Latitude']}
        plt.plot(model.parameters.index,
                 model.parameters[parameters[plt_type][0]],
                 marker=parameters[plt_type][1],
                 linestyle='',
                 color=parameters[plt_type][2],
                 label=parameters[plt_type][3])
        if len(model.parameters[parameters[plt_type][0]])-1 > 3:
            fit = parameter_fit(model.parameters.index, model.parameters[parameters[plt_type][0]], fit_args)
            plt.plot(fit['best_fit_x'], fit['best_fit_y'], '-', color=parameters[plt_type][2])
            plt.fill_between(fit['best_fit_x'], fit['sigma_bounds']['up'], fit['sigma_bounds']['low'],
                             color=parameters[plt_type][2], alpha=0.05)
        ylabel = parameters[plt_type][3] + ' [degrees]'

    plt.xlabel('Time [UT]')
    plt.ylabel(ylabel)
    if fit_args['type'] == 'polynomial':
        title = 'Event: ' + model.event_selected + ' | ' + fit_args['type'] + str(fit_args['order'])
    elif fit_args['type'] == 'spline':
        title = 'Event: ' + model.event_selected + ' | ' + fit_args['type'] + str(fit_args['order']) + ' (' + str(fit_args['smooth']) + ')'
    elif fit_args['type'] == 'custom':
        title = 'Event: ' + model.event_selected + ' | Funct: ' + fit_args['expression']

    plt.title(title)
    axis.xaxis.set_major_formatter(mdates.DateFormatter('%Y\n%b-%d\n%H:%M'))
    axis.minorticks_on()
    fig.autofmt_xdate(bottom=0, rotation=0, ha='center')
    plt.legend(loc='lower right')

    return fig


def parameter_fit(x, y, fit_args):
    '''
    Performs a polynomial or spline fit to a set of x, y parameters.
    '''
    xx = (mdates.date2num(x) - mdates.date2num(x[0]))
    xxx = np.linspace(xx.min(), xx.max(), 120)
    dd = mdates.num2date(xxx + mdates.date2num(x[0]))

    if fit_args['type'] == 'polynomial':
        # scipy.optimize.curve_fit and numpy.polyfit
        popt, pcov = np.polyfit(xx, y, fit_args['order'], full=False, cov=True)
        sigma = np.sqrt(np.diagonal(pcov))  # calculate sigma from covariance matrix
        best_fit = np.polyval(popt, xxx)
        sigma_bound_up = np.polyval((popt + sigma), xxx)
        sigma_bound_low = np.polyval((popt - sigma), xxx)
        fitting = {'popt': popt,
                   'pcov': pcov,
                   'sigma': sigma,
                   'best_fit_x_num': xxx,
                   'best_fit_x': dd,
                   'best_fit_y': best_fit,
                   'sigma_bounds': {'up': sigma_bound_up,
                                    'low': sigma_bound_low},
                   }
    elif fit_args['type'] == 'spline':
        spl = UnivariateSpline(xx, y, s=fit_args['smooth'], k=fit_args['order'])
        resid = y - spl(xx)  # true - prediction
        sigma = np.nanstd(resid, axis=0)  # sigma = std(res)
        best_fit = spl(xxx)
        sigma_bound_up = best_fit + sigma
        sigma_bound_low = best_fit - sigma

        sv_bound_up, sv_bound_low, sv_bound_dup, sv_bound_dlow = best_fit, best_fit, np.gradient(best_fit, xxx), np.gradient(best_fit, xxx)
        for i in range(2, 100):
            spl = UnivariateSpline(xx, y, s=i/100, k=fit_args['order'])
            sv_bound_up = np.maximum(sv_bound_up, spl(xxx))
            sv_bound_low = np.minimum(sv_bound_low, spl(xxx))
            sv_bound_dup = np.maximum(sv_bound_dup, np.gradient(spl(xxx), xxx))
            sv_bound_dlow = np.minimum(sv_bound_dlow, np.gradient(spl(xxx), xxx))

        fitting = {'spl': spl,
                   'sigma': sigma,
                   'best_fit_x_num': xxx,
                   'best_fit_x': dd,
                   'best_fit_y': best_fit,
                   'sigma_bounds': {'up': sigma_bound_up,
                                    'low': sigma_bound_low},
                   'sigv_bounds': {'up': sv_bound_up,
                                   'low': sv_bound_low,
                                   'dup': sv_bound_dup,
                                   'dlow': sv_bound_dlow},
                   }
    elif fit_args['type'] == 'custom':
        import numexpr
        from scipy.optimize import curve_fit

        expression = fit_args['expression']  # 'a * exp(-b * x) + c'

        def func(x, a, b, c):
            return numexpr.evaluate(expression)

        popt, pcov = curve_fit(func, xx, y, bounds=fit_args['bounds'], maxfev=2500)
        sigma = np.sqrt(np.diagonal(pcov))  # calculate sigma from covariance matrix
        best_fit = func(xxx, *popt)
        sigma_bound_up = func(xxx, *(popt + sigma))
        sigma_bound_low = func(xxx, *(popt - sigma))
        fitting = {'popt': popt,
                   'pcov': pcov,
                   'sigma': sigma,
                   'best_fit_x_num': xxx,
                   'best_fit_x': dd,
                   'best_fit_y': best_fit,
                   'sigma_bounds': {'up': sigma_bound_up,
                                    'low': sigma_bound_low},
                   }
    return fitting



"""
Test the geometrical models
"""

import astropy.units as u
import pytest
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose

from PyThea.geometrical_models import ellipsoid, sphere, spheroid


@pytest.fixture
def center():
    return SkyCoord(0*u.deg, 0*u.deg, 0*u.R_sun,
                    frame='heliographic_stonyhurst',
                    observer='earth',
                    obstime='2020/01/01T00:00:30')


def test_sphere_shape():
    """
    Tests if sphere produces the right matrix dimensions
    """
    x, y, z = sphere(n=20)
    assert (x.shape, y.shape, z.shape) == ((21, 21), (21, 21), (21, 21))


def test_sphere_isunit():
    """
    Tests that sphere is unit shpere
    """
    x, y, z = sphere(n=20)
    assert (x**2 + y**2 + z**2).all() == 1


def test_spheroid_pointlike(center):
    """
    Tests that spheroid is point-like when both semi-axes length is zero
    """
    sph = spheroid(center, 0.*u.R_sun, 0.*u.R_sun)
    assert (sph.coordinates == center).all()


def test_spheroid_isunit(center):
    """
    Tests that spheroid is unit sphere when both semi-axes are equal to one
    """
    sph = spheroid(center, 1.*u.R_sun, 1.*u.R_sun)
    assert_quantity_allclose(sph.coordinates.radius, 1.*u.R_sun, atol=5e-7*u.R_sun)


# TODO def test_spheroid_isunit_afrer_rotation(center):


def test_spheroid_apex_location():
    """
    Tests that the spheroid apex location is correct
    """
    center_ = SkyCoord(0*u.deg, 0*u.deg, 1*u.R_sun,
                       frame='heliographic_stonyhurst',
                       observer='earth',
                       obstime='2020/01/01T00:00:30')
    apex_ = SkyCoord(0*u.deg, 0*u.deg, 2*u.R_sun,
                       frame='heliographic_stonyhurst',
                       observer='earth',
                       obstime='2020/01/01T00:00:30')
    sph = spheroid(center_, 1.*u.R_sun, 1.*u.R_sun)
    assert_quantity_allclose(sph.apex.radius, apex_.radius, atol=5e-7*u.R_sun)


# TODO @pytest.mark.parametrize('num, expected', [(-1, 1), (0, 2), (2, 4)])
def test_spheroid_parameters_conversions():
    """
    Tests that the spheroid parameters conversion is correct
    """
    he, ep, ka = spheroid.hek_from_rab(1*u.R_sun, 2*u.R_sun, 3*u.R_sun)
    r, a, b = spheroid.rab_from_hek(he, ep, ka)
    he_, ep_, ka_ = spheroid.hek_from_rab(r, a, b)
    assert_quantity_allclose(he, he_, atol=5e-7*u.R_sun)
    assert_quantity_allclose(ep, ep_, atol=5e-7)
    assert_quantity_allclose(ka, ka_, atol=5e-7)


def test_ellipsoid_semiaxis_alpha(center):
    """
    Tests that the ellipsoid alpha parameter is calculated correct
    """
    ell = ellipsoid(center, 1.*u.R_sun, 1.*u.R_sun, 1.*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(ell.alpha, 1, atol=5e-7)
    ell = ellipsoid(center, 1.*u.R_sun, 1.*u.R_sun, 2.*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(ell.alpha, 0.5, atol=5e-7)
    ell = ellipsoid(center, 1.*u.R_sun, 2.*u.R_sun, 1.*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(ell.alpha, 2, atol=5e-7)


def test_ellipsoid_parameters_conversions():
    """
    Tests that the ellipsoid parameters conversion is correct
    """
    he, ep, ka, al = ellipsoid.heka_from_rabc(1*u.R_sun, 2*u.R_sun, 3*u.R_sun, 4*u.R_sun)
    r, a, b, c = ellipsoid.rabc_from_heka(he, ep, ka, al)
    he_, ep_, ka_, al_ = ellipsoid.heka_from_rabc(r, a, b, c)
    assert_quantity_allclose(he, he_, atol=5e-7*u.R_sun)
    assert_quantity_allclose(ep, ep_, atol=5e-7)
    assert_quantity_allclose(ka, ka_, atol=5e-7)
    assert_quantity_allclose(al, al_, atol=5e-7)


def test_ellipsoid_unitsphere_equivalency(center):
    """
    Tests that an ellipsoid with all the semi-axis with length one is unit sphere
    """
    ell = ellipsoid(center, 1.*u.R_sun, 1.*u.R_sun, 1.*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(ell.coordinates.radius, 1.*u.R_sun, atol=5e-7*u.R_sun)


def test_spheroid_ellipsoid_equivalency(center):
    """
    Tests that an ellipsoid and a spheroid are equivalent
    """
    sph = spheroid(center, 1.5*u.R_sun, 1.5*u.R_sun)
    ell = ellipsoid(center, 1.5*u.R_sun, 1.5*u.R_sun, 1.5*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(sph.coordinates.radius, ell.coordinates.radius, atol=5e-7*u.R_sun)
    sph = spheroid(center, 2*u.R_sun, 1.5*u.R_sun)
    ell = ellipsoid(center, 2*u.R_sun, 1.5*u.R_sun, 1.5*u.R_sun, 0.*u.degree)
    assert_quantity_allclose(sph.coordinates.radius, ell.coordinates.radius, atol=5e-7*u.R_sun)

# TODO Compare the stony and carr center values



import numpy as np
import sunpy.map

from sunpy_dev.map import maputils

__all__ = ['cor_polariz']


def cor_polariz(map_sequence):
    """
    Returns the total brightness image from three polarized images.

    Parameters
    ----------
    smap : `~sunpy.map.GenericMap`
        A SunPy map.

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.

    Note
    -------
    This script is partially based on cor_polariz.pro
    """
    sequence_polar_0 = [tmap for tmap in map_sequence if tmap.meta['polar'] == 0]
    sequence_polar_120 = [tmap for tmap in map_sequence if tmap.meta['polar'] == 120]
    sequence_polar_240 = [tmap for tmap in map_sequence if tmap.meta['polar'] == 240]

    # Calculate Mueller matrix
    s, d = 0.5, 0.5
    ang_0, ang_120, ang_240 = 0, 120, 240

    m = [[s, d*np.cos(2*np.deg2rad(ang_0)), d*np.sin(2*np.deg2rad(ang_0))],
         [s, d*np.cos(2*np.deg2rad(ang_120)), d*np.sin(2*np.deg2rad(ang_120))],
         [s, d*np.cos(2*np.deg2rad(ang_240)), d*np.sin(2*np.deg2rad(ang_240))]]
    m = np.linalg.inv(m)

    # Calculate the Stokes parameters for incident intensity
    I = m[0][0] + m[0][1] + m[0][2]
    # Q = m[1][0] + m[1][1] + m[1][2]
    # U = m[2][0] + m[2][1] + m[2][2]

    smap = []
    for map_0 in sequence_polar_0:
        map_120 = maputils.get_closest(sequence_polar_120, map_0.date)
        map_240 = maputils.get_closest(sequence_polar_240, map_0.date)

        map_0 = maputils.normalize_exposure(map_0)
        map_120 = maputils.normalize_exposure(map_120)
        map_240 = maputils.normalize_exposure(map_240)

        # Calculate the Stokes parameters for incident intensity
        I = map_0.data * m[0][0] + map_120.data * m[0][1] + map_240.data * m[0][2]
        # Q = map_0.data * m[1][0] + map_120.data * m[1][1] + map_240.data * m[1][2]
        # U = map_0.data * m[2][0] + map_120.data * m[2][1] + map_240.data * m[2][2]

        t = [map_0.date, map_120.date, map_240.date]
        obs_time = min(t) + (max(t)-min(t))/2
        crota = np.mean([map_0.meta['crota'], map_120.meta['crota'], map_240.meta['crota']])

        new_map_meta = map_120.meta
        # TODO: Check if any other keys need to change
        new_map_meta['date-obs'] = obs_time.strftime('%Y-%m-%dT%H:%M:%S.%f')
        new_map_meta['polar'] = '1001'
        new_map_meta['crval1'] = np.mean([map_0.meta['crval1'], map_120.meta['crval1'], map_240.meta['crval1']])
        new_map_meta['crval2'] = np.mean([map_0.meta['crval2'], map_120.meta['crval2'], map_240.meta['crval2']])
        new_map_meta['crpix1'] = np.mean([map_0.meta['crpix1'], map_120.meta['crpix1'], map_240.meta['crpix1']])
        new_map_meta['crpix2'] = np.mean([map_0.meta['crpix2'], map_120.meta['crpix2'], map_240.meta['crpix2']])
        new_map_meta['crota'] = crota
        new_map_meta['PC1_1'] = np.cos(np.deg2rad(crota))
        new_map_meta['PC1_2'] = -np.sin(np.deg2rad(crota))
        new_map_meta['PC2_1'] = np.sin(np.deg2rad(crota))
        new_map_meta['PC2_2'] = np.cos(np.deg2rad(crota))

        map_ = sunpy.map.Map(I, new_map_meta)
        smap.append(map_)

    return sunpy.map.Map(smap, sequence=True)



import astropy.units as u
from sunpy.coordinates import frames


def load_or_delete_fittings(st):
    selected_row = str(st.session_state.fitting_select)
    dataframe = st.session_state.model_fittings.parameters
    if st.session_state.fit_action == 'Select':
        pass
    elif st.session_state.fit_action == 'Load':
        if st.session_state.coord_system == 'HGS':
            st.session_state.longit = float(dataframe.loc[selected_row, 'hgln'])
            st.session_state.latitu = float(dataframe.loc[selected_row, 'hglt'])
        elif st.session_state.coord_system == 'HGC':
            st.session_state.longit = float(dataframe.loc[selected_row, 'crln'])
            st.session_state.latitu = float(dataframe.loc[selected_row, 'crlt'])

        if st.session_state.model_fittings.geometrical_model == 'Spheroid':
            keys = ['height', 'kappa', 'epsilon', 'rcenter', 'radaxis', 'orthoaxis1']
        elif st.session_state.model_fittings.geometrical_model == 'Ellipsoid':
            keys = ['height', 'kappa', 'epsilon', 'alpha', 'rcenter', 'radaxis', 'orthoaxis1', 'orthoaxis2', 'tilt']
        elif st.session_state.model_fittings.geometrical_model == 'GCS':
            keys = ['height', 'alpha', 'kappa', 'tilt']

        for key in keys:
            if key in st.session_state:
                # del st.session_state[key]
                st.session_state[key] = float(dataframe.loc[selected_row, key])
        del st.session_state.fit_action
    elif st.session_state.fit_action == 'Delete':
        st.session_state.model_fittings.parameters = dataframe.drop(st.session_state.fitting_select)
        del st.session_state.fitting_select
        if len(st.session_state.model_fittings.parameters) < 1:
            del st.session_state.model_fittings
        del st.session_state.fit_action


def change_long_lat_sliders(st):
    if st.session_state.coord_system == 'HGS':
        center_ = st.session_state.center.transform_to(frames.HeliographicStonyhurst)
    elif st.session_state.coord_system == 'HGC':
        center_ = st.session_state.center.transform_to(frames.HeliographicCarrington)
    st.session_state.longit = float(center_.lon.to_value(u.degree))
    st.session_state.latitu = float(center_.lat.to_value(u.degree))



"""
A place for the more generic tests that do not fit in the other categories

Note
----
Before you start the test the PyThea pkg should be in the python path
export PYTHONPATH="${PYTHONPATH}:{top_level_dir_that_pythea_lives}/PyThea"
"""



'''
 A dictionary of the selected imagers
 How to use this:

'''
import astropy.units as u
from sunpy.net import attrs as a

imager_dict = dict.fromkeys(['LC2'], [(a.Instrument.lasco, a.Detector.c2),
                                      {'dimensions': (1024*u.pixel, 1024*u.pixel), 'polar': 'Clear', 'superpixel': 2}, 'LASCO-C2'])
imager_dict.update(dict.fromkeys(['LC3'], [(a.Instrument.lasco, a.Detector.c3),
                                           {'dimensions': (1024*u.pixel, 1024*u.pixel), 'polar': 'Clear', 'superpixel': 2}, 'LASCO-C3']))
imager_dict.update(dict.fromkeys(['AIA'], [(a.Instrument.aia, a.Wavelength(19.3 * u.nm), a.Sample(1*u.minute)),
                                           {'dimensions': (4096*u.pixel, 4096*u.pixel), 'superpixel': 8, 'exposure': 1.90}, 'SDO-AIA']))
imager_dict.update(dict.fromkeys(['COR2A'], [(a.Source('STEREO_A'), a.Detector.cor2),
                                             {'dimensions': (2048*u.pixel, 2048*u.pixel), 'polar': 1001, 'superpixel': 4}, 'STA-COR2']))
imager_dict.update(dict.fromkeys(['COR2B'], [(a.Source('STEREO_B'), a.Detector.cor2),
                                             {'dimensions': (2048*u.pixel, 2048*u.pixel), 'polar': 1001, 'superpixel': 4}, 'STB-COR2']))
imager_dict.update(dict.fromkeys(['EUVIA'], [(a.Source('STEREO_A'), a.Detector.euvi, a.Wavelength(19.5 * u.nm)),
                                             {'dimensions': (2048*u.pixel, 2048*u.pixel), 'superpixel': 4}, 'STA-EUVI']))
imager_dict.update(dict.fromkeys(['EUVIB'], [(a.Source('STEREO_B'), a.Detector.euvi, a.Wavelength(19.5 * u.nm)),
                                             {'dimensions': (2048*u.pixel, 2048*u.pixel), 'superpixel': 4}, 'STB-EUVI']))
imager_dict.update(dict.fromkeys(['COR1A'], [(a.Source('STEREO_A'), a.Detector.cor1),
                                             {'dimensions': (512*u.pixel, 512*u.pixel)}, 'STA-COR1']))  # 'polar':0
imager_dict.update(dict.fromkeys(['COR1B'], [(a.Source('STEREO_B'), a.Detector.cor1),
                                             {'dimensions': (512*u.pixel, 512*u.pixel)}, 'STB-COR1']))  # 'polar':0
imager_dict.update(dict.fromkeys(['HI1A'], [(a.Source('STEREO_A'), a.Detector.hi1),
                                            {'dimensions': (1024*u.pixel, 1024*u.pixel), 'superpixel': 2}, 'STA-HI1']))
imager_dict.update(dict.fromkeys(['HI1B'], [(a.Source('STEREO_B'), a.Detector.hi1),
                                            {'dimensions': (1024*u.pixel, 1024*u.pixel), 'superpixel': 2}, 'STB-HI1']))
imager_dict.update(dict.fromkeys(['HI2A'], [(a.Source('STEREO_A'), a.Detector.hi2),
                                            {'dimensions': (1024*u.pixel, 1024*u.pixel), 'superpixel': 2}, 'STA-HI2']))
imager_dict.update(dict.fromkeys(['HI2B'], [(a.Source('STEREO_B'), a.Detector.hi2),
                                            {'dimensions': (1024*u.pixel, 1024*u.pixel)}, 'STB-HI2']))



"""
    PyThea: A software package to reconstruct the 3D structure of CMEs and
    shock waves using multi-viewpoint remote-sensing observations.
    Copyright (C) 2021  Athanasios Kouloumvakos

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import operator
import re

import astropy.units as u
import numpy as np
import pandas as pd
import seaborn as sns
from astropy.coordinates import (CartesianRepresentation, Distance, SkyCoord,
                                 SphericalRepresentation, concatenate)
from numpy.linalg import norm
from scipy.spatial.transform import Rotation
from sunpy.coordinates import frames


def sphere(n=20):
    """
    Returns the (x,y,z) coordinates of a unit sphere centred at the origin.
    The coordinates are (n+1)-by-(n+1) matrices.
    """

    theta = np.linspace(-np.pi, np.pi, n+1)
    phi = np.linspace(-1*np.pi/2, 1 * np.pi/2, n+1)
    cosphi = np.cos(phi)
    sintheta = np.sin(theta)

    x =  np.outer(cosphi, np.cos(theta))
    y =  np.outer(cosphi, sintheta)
    z =  np.outer(np.sin(phi), np.ones_like(theta))

    return [x, y, z]


class spheroid:
    '''
    A class for the spheroid model.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The coordinates of the spheroid center.
    rcenter : `~astropy.units.Quantity`
        The radial distance of the spheroid center from the solar center.
    radaxis : `~astropy.units.Quantity`
        The spheroid's first (radial) semi-axis length.
    orthoaxis1 : `~astropy.units.Quantity`
        The spheroid's second semi-axis length (orthogonal axis wrt the radial axis).
    height : `~astropy.units.Quantity`
        The radial distance of the spheroid apex from the solar center (e.g., rcenter+radaxis).
    kappa : `~astropy.units.Quantity`
        Spheroid self-similar constant, defined as the ratio of the apex height to the length
        of the orthoaxis1. (This is proportional to the aspect ratio of the spheroid)
    epsilon : `~astropy.units.Quantity`
        The eccentricity of the spheroid.

    Notes
    -----
    Details about spheroids can be found here: https://en.wikipedia.org/wiki/Spheroid
    The second and third semi-axis are equal in the spheroid. If all the axis are equal we have a sphere.
    Tilt defaults to zero, since any rotation along the radial axis is trivial for the spheroid.
    '''
    @u.quantity_input
    def __init__(self, center, radaxis: (u.R_sun), orthoaxis1: (u.R_sun), n=40):
        self.center = center.transform_to(frames.HeliographicStonyhurst)  # This is the SkyCoord of the ellipsoid center
        self.rcenter = center.radius
        self.radaxis = radaxis
        self.orthoaxis1 = orthoaxis1

        self.height = radaxis + self.rcenter
        self.kappa = orthoaxis1 / (self.height-1.*u.R_sun)

        if radaxis < orthoaxis1:
            self.epsilon = -1. * np.sqrt(1. - (radaxis/orthoaxis1)**2)
        elif radaxis > orthoaxis1:
            self.epsilon = 1. * np.sqrt(1. - (orthoaxis1/radaxis)**2)
        elif radaxis == orthoaxis1:
            self.epsilon = 0.

        self.tilt = 0 * u.degree
        self.n = n

    @property
    def coordinates(self):
        """
        Returns the coordinates of the spheroid cloud of points as `~astropy.coordinates.SkyCoord`.
        """
        [x__, y__, z__] = sphere(self.n)

        x_ = self.radaxis * x__ + self.rcenter
        y_ = self.orthoaxis1 * y__
        z_ = self.orthoaxis1 * z__

        x, y, z = self.rotate(x_, y_, z_)

        return SkyCoord(CartesianRepresentation(x, y, z),
                        frame=frames.HeliographicStonyhurst,
                        observer=self.center.observer,
                        obstime=self.center.obstime)

    def intersecting_curve(self):
        """
        Returns the coordinates of intersection of the spheroid with a unit sphere
         centred at the origin as `~astropy.coordinates.SkyCoord`.
         More information here: https://mathworld.wolfram.com/Sphere-SphereIntersection.html
        """
        d = self.rcenter
        r = self.radaxis
        R = 1 * u.R_sun

        step = 40
        theta = np.linspace(0, 2*np.pi, step)
        x_ = np.ones_like(theta) * (d**2 - r**2 + R**2) / (2 * d)
        # y_**2 - z_**2 = R**2 - x_**2
        alpha = np.sqrt(4 * d**2 * R**2 - (d**2 - r**2 + R**2)**2) / (2 * d)
        y_ = alpha * np.sin(theta)
        z_ = alpha * np.cos(theta)
        x, y, z = self.rotate(x_, y_, z_)

        return SkyCoord(CartesianRepresentation(x, y, z),
                        frame=frames.HeliographicStonyhurst,
                        observer=self.center.observer,
                        obstime=self.center.obstime)

    @staticmethod
    @u.quantity_input
    def hek_from_rab(rc: (u.R_sun), radaxis: (u.R_sun), orthoaxis1: (u.R_sun)):
        '''
        Returns the semi-axis length from the height, epsilon, and kappa parameters
        '''
        height = rc + radaxis
        kappa = orthoaxis1 / (height - 1 * u.R_sun)

        if radaxis < orthoaxis1:
            epsilon = -1. * np.sqrt(1 - (radaxis / orthoaxis1)**2)
        elif radaxis > orthoaxis1:
            epsilon = np.sqrt(1. - (orthoaxis1 / radaxis)**2)
        elif radaxis == orthoaxis1:
            epsilon = 0.
        else:
            height, kappa, epsilon = np.nan * u.R_sun, np.nan, np.nan

        return height, epsilon, kappa

    @staticmethod
    @u.quantity_input
    def rab_from_hek(height: (u.R_sun), epsilon, kappa):
        '''
        Returns the height, epsilon, and kappa parameters from the semi-axis length
        '''
        s = kappa * (height - 1. * u.R_sun)
        if epsilon < 0:
            d = s * np.sqrt(1. - epsilon**2)
        elif epsilon > 0:
            d = s / np.sqrt(1. - epsilon**2)
        elif epsilon == 0:
            d = s
        else:
            rc, radaxis, orthoaxis1 = np.nan * u.R_sun, np.nan * u.R_sun, np.nan * u.R_sun
            return rc, radaxis, orthoaxis1
        radaxis = d
        orthoaxis1 = s
        rc = height - radaxis

        return rc, radaxis, orthoaxis1

    @property
    def apex(self):
        """
        Returns the coordinates of the spheroid apex as `~astropy.coordinates.SkyCoord`.
        """
        rapex = self.center.radius + self.radaxis
        Spher_rep = SphericalRepresentation(self.center.lon, self.center.lat,
                                            Distance(np.abs(rapex)))
        return SkyCoord(Spher_rep.to_cartesian(),
                        frame=frames.HeliographicStonyhurst,
                        obstime=self.center.obstime,
                        observer=self.center.observer)

    @property
    def base(self):
        """
        Returns the coordinates of the spheroid base as `~astropy.coordinates.SkyCoord`.
        """
        rbase = (-self.center.radius + self.radaxis)
        Spher_rep = SphericalRepresentation(self.center.lon, self.center.lat,
                                            Distance(np.abs(rbase)))
        return SkyCoord(-np.sign(rbase) * Spher_rep.to_cartesian(),
                        frame=frames.HeliographicStonyhurst,
                        obstime=self.center.obstime,
                        observer=self.center.observer)

    def rotate(self, x_, y_, z_):
        """
        Rotates the (x,y,z) coordinates from the geometrical model's coordinate system
         to the Heliospheric coordinates.

        Returns the coordinates as `~astropy.coordinates.SkyCoord`.
        """
        Longc = self.center.lon.to_value(u.rad)
        Latc  = self.center.lat.to_value(u.rad)
        tilt = self.tilt.to_value(u.rad)

        v = np.transpose([x_.flatten(), y_.flatten(), z_.flatten()])
        v = Rotation.from_euler('xyz', [tilt, -Latc, Longc]).apply(v)

        x = np.reshape(v[:, 0], x_.shape) * x_.unit
        y = np.reshape(v[:, 1], y_.shape) * y_.unit
        z = np.reshape(v[:, 2], z_.shape) * z_.unit
        return x, y, z

    def plot(self, axis, mode='Skeleton', only_surface=False):
        '''
        Plots the spheroid model at the provided axes
        '''
        if mode == 'Full':
            axis.plot_coord(self.coordinates, color='red', linestyle='-', linewidth=0.2)
        elif (mode == 'Skeleton') and (only_surface is False):
            lw = 1.
            n = self.n
            palete = sns.color_palette('colorblind')

            # Plot the part of the spheroid mesh but filter lines bellow the solar surface
            axis.plot_coord(self.coordinates[:, int(n/2)], color=palete[0], linestyle='-', linewidth=lw)
            my_plot_coord(self.coordinates[:, 0], axis, color=palete[6], linestyle='-', linewidth=lw)
            axis.plot_coord(self.coordinates[0:int(n/2)+1, int(n/4)], color=palete[1], linestyle='-', linewidth=lw)
            axis.plot_coord(self.coordinates[int(n/2):n+1, int(n/4)], color=palete[1], linestyle='-', linewidth=lw)
            axis.plot_coord(self.coordinates[0:int(n/2)+1, int(3*n/4)], color=palete[1], linestyle='-', linewidth=lw)
            axis.plot_coord(self.coordinates[int(n/2):n+1, int(3*n/4)], color=palete[1], linestyle='-', linewidth=lw)
            my_plot_coord(self.coordinates[int(n/2), 0:int(n/4)+1], axis, color=palete[2], linestyle='-', linewidth=lw)
            my_plot_coord(self.coordinates[int(n/2), int(3*n/4):n+1], axis, color=palete[2], linestyle='-', linewidth=lw)
            axis.plot_coord(self.coordinates[int(n/2), int(n/4):int(3*n/4)+1], color=palete[3], linestyle='-', linewidth=lw)

        if (mode == 'Skeleton') or (only_surface is True):
            # Plot the part of the intersecting curve at the solar surface
            d_val = self.rcenter - self.radaxis
            if d_val > -1*u.R_sun and d_val < 1*u.R_sun:
                coords = self.intersecting_curve()
                axes_frame = axis._transform_pixel2world.frame_out
                coord_in_axes = coords.transform_to(axes_frame)
                rsun = getattr(axes_frame, 'rsun', u.R_sun)
                reference_distance = np.sqrt(axes_frame.observer.radius**2 - rsun**2)
                is_visible = coord_in_axes.spherical.distance <= reference_distance
                if np.any(is_visible):
                    axis.plot_coord(coords[is_visible], linestyle='', marker='+', markersize=3, color=(0.9, 0.24, 0.38, 1))
                    # coord_=coords[is_visible]
                    # s=''.join('X' if p else 'O' for p in np.diff(np.diff(coord_.lat))<5*u.deg)
                    # clist = list([m.span()[0], abs(operator.sub(*m.span()))] for m in re.finditer('X+', s))
                    # for c in clist:
                    #   axis.plot_coord(coord_[c[0]:(c[0]+c[1])], linestyle='-', linewidth=0.8, color=(0, 0, 0, 1))
                if np.any(~is_visible):
                    axis.plot_coord(coords[~is_visible], linestyle='', marker='+', markersize=3, color=(0, 0, 0, 1))
                    # coord_ = coords[~is_visible]
                    # s = ''.join('X' if p else 'O' for p in np.diff(np.diff(coord_.lat))<5*u.deg)
                    # clist = list([m.span()[0], abs(operator.sub(*m.span()))] for m in re.finditer('X+', s))
                    # for c in clist:
                    #    axis.plot_coord(coord_[c[0]:(c[0]+c[1])], linestyle='--', linewidth=0.8, color=(0, 0, 0, 1))

        axis.plot_coord(self.apex, marker='o', color=(0, 0, 0, 1))
        axis.plot_coord(self.base, marker='x', color=(0, 0, 0, 1))
        axis.plot_coord(self.center, marker='+', color=(0, 0, 0, 1))
        axis.plot_coord(concatenate((self.apex, self.base)), linestyle='-', linewidth=0.5, color=(0, 0, 0, 1))

    def to_dataframe(self):
        """
        Returns the spheroid parameters as `~pandas.DataFrame`.
        """
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        data_dict = {
            'hgln': self.center.lon.to_value(u.degree),
            'hglt': self.center.lat.to_value(u.degree),
            'crln': center_.lon.to_value(u.degree),
            'crlt': center_.lat.to_value(u.degree),
            'rcenter': self.rcenter.to_value(u.R_sun),
            'radaxis': self.radaxis.to_value(u.R_sun),
            'orthoaxis1': self.orthoaxis1.to_value(u.R_sun),
            'height': self.height.to_value(u.R_sun),
            'kappa': self.kappa,
            'epsilon': self.epsilon,
        }

        return pd.DataFrame(data_dict, index=[self.center.obstime.to_datetime()])

    def __str__(self):
        '''
        Returns the Spheroid object and parameters in a printable sting array.
        '''
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        output = '<Spheroid object \n'
        output += 'HGLN = %3.2f degrees \n'%self.center.lon.to_value(u.degree)
        output += 'HGLT = %3.2f degrees \n'%self.center.lat.to_value(u.degree)
        output += 'CRLN = %3.2f degrees \n'%center_.lon.to_value(u.degree)
        output += 'CRLT = %3.2f degrees \n'%center_.lat.to_value(u.degree)
        output += 'rcenter = %3.2f Rsun \n'%self.rcenter.to_value(u.R_sun)
        output += 'radaxis = %3.2f Rsun \n'%self.radaxis.to_value(u.R_sun)
        output += 'height = %3.2f Rsun \n'%self.radaxis.to_value(u.R_sun)
        output += 'kappa = %3.2f \n'%self.kappa
        output += 'epsilon = %3.2f '%self.epsilon
        output += '>'
        return output


class ellipsoid(spheroid):
    '''
    A class for the ellipsoid model.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The coordinates of the ellipsoid center.
    rcenter : `~astropy.units.Quantity`
        The radial distance of the ellipsoid center from the solar center.
    radaxis : `~astropy.units.Quantity`
        The ellipsoid's first (radial) semi-axis length.
    orthoaxis1 : `~astropy.units.Quantity`
        The ellipsoid's second semi-axis length (orthogonal axis wrt the radial axis).
    orthoaxis2 : `~astropy.units.Quantity`
        The ellipsoid's third semi-axis axis length (orthogonal axis wrt the radial axis).
    height : `~astropy.units.Quantity`
        The radial distance of the ellipsoid apex from the solar center (e.g., rcenter+radaxis).
    kappa : `~astropy.units.Quantity`
        Ellipsoid self-similar constant, defined as the ratio of the apex height to the length
        of the orthoaxis1. (This is proportional to the first aspect ratio of the ellipsoid)
    alpha : `~astropy.units.Quantity`
        The ellipsoid's second aspect ratio, namely the ratio between the second and the third
         semi-axis.
    epsilon : `~astropy.units.Quantity`
        The eccentricity of the ellipsoid.
    tilt : `~astropy.units.Quantity`
        The tilt angle of the ellipsoid. The axis of rotation is the first (radial) semi-axis.

    Notes
    -----
    Details about ellipsoids can be found here: https://en.wikipedia.org/wiki/Ellipsoid
    '''
    def __init__(self, center, radaxis: (u.R_sun), orthoaxis1: (u.R_sun), orthoaxis2: (u.R_sun), tilt: (u.degree), n=40):
        super().__init__(center, radaxis, orthoaxis1, n)

        self.orthoaxis2 = orthoaxis2
        self.tilt = tilt
        self.alpha = orthoaxis1 / orthoaxis2

    def intersecting_curve(self):
        """
        Returns the coordinates of intersection of the ellipsoid with a unit sphere
         centred at the origin as `~astropy.coordinates.SkyCoord`.
        We use vedo package for this calculation.
        """
        from vedo import Ellipsoid, Sphere
        sph = Sphere(pos=(0, 0, 0), r=1)
        ell = Ellipsoid(pos=(self.rcenter.to_value(1*u.R_sun), 0, 0),
                        axis1=(2*self.radaxis.to_value(1*u.R_sun), 0, 0),
                        axis2=(0, 2*self.orthoaxis1.to_value(1*u.R_sun), 0),
                        axis3=(0, 0, 2*self.orthoaxis2.to_value(1*u.R_sun)))
        sic = sph.intersectWith(ell)
        poi = sic.points()
        x, y, z = self.rotate(poi[:, 0]*u.R_sun, poi[:, 1]*u.R_sun, poi[:, 2]*u.R_sun)

        return SkyCoord(CartesianRepresentation(x, y, z),
                        frame=frames.HeliographicStonyhurst,
                        observer=self.center.observer,
                        obstime=self.center.obstime)

    @property
    def coordinates(self):
        """
        Returns the coordinates of the ellipsoid cloud of points as `~astropy.coordinates.SkyCoord`.
        """
        [x__, y__, z__] = sphere(self.n)

        x_ = self.radaxis * x__ + self.rcenter
        y_ = self.orthoaxis1 * y__
        z_ = self.orthoaxis2 * z__

        x, y, z = self.rotate(x_, y_, z_)

        return SkyCoord(CartesianRepresentation(x, y, z),
                        frame=frames.HeliographicStonyhurst,
                        observer=self.center.observer,
                        obstime=self.center.obstime)

    def to_dataframe(self):
        """
        Returns the ellipsoid parameters as `~pandas.DataFrame`.
        """
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        data_dict = {
            'hgln': self.center.lon.to_value(u.degree),
            'hglt': self.center.lat.to_value(u.degree),
            'crln': center_.lon.to_value(u.degree),
            'crlt': center_.lat.to_value(u.degree),
            'rcenter': self.rcenter.to_value(u.R_sun),
            'radaxis': self.radaxis.to_value(u.R_sun),
            'orthoaxis1': self.orthoaxis1.to_value(u.R_sun),
            'orthoaxis2': self.orthoaxis2.to_value(u.R_sun),
            'tilt': self.tilt.to_value(u.degree),
            'height': self.height.to_value(u.R_sun),
            'kappa': self.kappa,
            'epsilon': self.epsilon,
            'alpha': self.alpha,
        }

        return pd.DataFrame(data_dict, index=[self.center.obstime.to_datetime()])

    @staticmethod
    @u.quantity_input
    def heka_from_rabc(rc: (u.R_sun), radaxis: (u.R_sun), orthoaxis1: (u.R_sun), orthoaxis2: (u.R_sun)):
        '''
        Returns the height, epsilon, kappa, and alpha parameters from the semi-axis length
        '''
        height = rc + radaxis
        kappa = orthoaxis1 / (height - 1 * u.R_sun)
        alpha = orthoaxis1 / orthoaxis2

        if radaxis < orthoaxis1:
            epsilon = -1. * np.sqrt(1 - (radaxis/orthoaxis1)**2)
        elif radaxis > orthoaxis1:
            epsilon = np.sqrt(1. - (orthoaxis1/radaxis)**2)
        elif radaxis == orthoaxis1:
            epsilon = 0.
        else:
            height, kappa, epsilon, alpha = np.nan * u.R_sun, np.nan, np.nan, np.nan

        return height, epsilon, kappa, alpha

    @staticmethod
    @u.quantity_input
    def rabc_from_heka(height: (u.R_sun), epsilon, kappa, alpha):
        '''
        Returns the semi-axis length from the height, epsilon, kappa, and alpha parameters
        '''
        s = kappa * (height - 1. * u.R_sun)
        if epsilon < 0:
            d = s * np.sqrt(1. - epsilon**2)
        elif epsilon > 0:
            d = s / np.sqrt(1. - epsilon**2)
        elif epsilon == 0:
            d = s
        else:
            rc, radaxis, orthoaxis1, orthoaxis2 = np.nan * u.R_sun, np.nan * u.R_sun, np.nan * u.R_sun, np.nan * u.R_sun
            return rc, radaxis, orthoaxis1, orthoaxis2
        radaxis = d
        orthoaxis1 = s
        orthoaxis2 = s / alpha
        rc = height - radaxis

        return rc, radaxis, orthoaxis1, orthoaxis2

    def __str__(self):
        '''
        Returns the Ellipsoid object and parameters in a printable sting array.
        '''
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        output = '<Ellipsoid object \n'
        output += 'HGLN = %3.2f degrees \n'%self.center.lon.to_value(u.degree)
        output += 'HGLT = %3.2f degrees \n'%self.center.lat.to_value(u.degree)
        output += 'CRLN = %3.2f degrees \n'%center_.lon.to_value(u.degree)
        output += 'CRLT = %3.2f degrees \n'%center_.lat.to_value(u.degree)
        output += 'rcenter = %3.2f Rsun \n'%self.rcenter.to_value(u.R_sun)
        output += 'radaxis = %3.2f Rsun \n'%self.radaxis.to_value(u.R_sun)
        output += 'orthoaxis1 = %3.2f Rsun \n'%self.orthoaxis1.to_value(u.R_sun)
        output += 'orthoaxis2 = %3.2f Rsun \n'%self.orthoaxis2.to_value(u.R_sun)
        output += 'height = %3.2f Rsun \n'%self.radaxis.to_value(u.R_sun)
        output += 'kappa = %3.2f \n'%self.kappa
        output += 'alpha = %3.2f \n'%self.alpha
        output += 'epsilon = %3.2f '%self.epsilon
        output += '>'
        return output


class gcs():
    """
    A class of the GCS CME model.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The coordinates of the GCS center (defined at the apex circle center).
    rcenter : `~astropy.units.Quantity`
        The radial distance of the GCS center from the solar center.
    height : `~astropy.units.Quantity`
        The radial distance of the GCS apex from the solar center (e.g., rcenter+rapex).
    kappa : `~astropy.units.Quantity`
        Aspect ratio of the GCS.
    alpha : `~astropy.units.Quantity`
        Half angular width of the GCS model.
    tilt : `~astropy.units.Quantity`
        The tilt angle of the GCS.

    Notes
    -----
    This is based on IDL script in here: https://hesperia.gsfc.nasa.gov/ssw/stereo/secchi/idl/scraytrace/shellskeleton.pro
    and with some additions from here: https://gitlab.physik.uni-kiel.de/ET/gcs_python/-/blob/master/gcs/geometry.py
    A full description of the GCS model is given in Thernisien et al. (2006) and Thernisien (2011).
    """
    def __init__(self, center, height: (u.R_sun), alpha: (u.degree), kappa, tilt: (u.degree),
                 nbvertsl=10, nbvertcirc=20, nbvertcircshell=90):
        self.center = center.transform_to(frames.HeliographicStonyhurst)  # This is the SkyCoord of the center of circle at the apex
        self.rcenter = center.radius  # rcenter = height-rapex
        self.height = height  # GCS height in apex
        self.alpha = alpha  # GCS CME width (in half angle)
        self.kappa = kappa  # Aspect ratio
        self.tilt = tilt  # Tilt angle

        self.h = self.h_()  # Height of the cone (CME legs)
        self.rapex = self.rapex_()  # Cross-section radius at the apex

        self.nbvertsl = nbvertsl
        self.nbvertcirc = nbvertcirc
        self.nbvertcircshell = nbvertcircshell

    @staticmethod
    @u.quantity_input
    def rcenter_(height: (u.R_sun), alpha: (u.degree), kappa):
        """
        Returns the distance of the GCS center from the solar center (OC1 in T2011, see eq. 20).
        """
        h = height * (1. - kappa) * np.cos(alpha) / (1. + np.sin(alpha))
        rapex = kappa * (h / np.cos(alpha) +  h * np.tan(alpha)) / (1 - kappa ** 2)
        return height-rapex

    def h_(self):
        """
        Returns the full height (h) of the cone (OD distance in T2011, see eq. 2).
        We use eq. 3 in T2006 to derive the h from height, kappa, and alpha parameters
        """
        h = self.height * (1. - self.kappa) * np.cos(self.alpha) / (1. + np.sin(self.alpha))
        return h

    def rapex_(self):
        """
        Returns the cross-section radius at the apex (see eq. 29 in T2011).
        """
        rapex = self.kappa * (self.h_() / np.cos(self.alpha) +  self.h_() * np.tan(self.alpha)) / (1 - self.kappa ** 2)
        return rapex

    @property
    def coordinates(self):
        """
        Returns the coordinates of the ellipsoid cloud of points as `~astropy.coordinates.SkyCoord`.
        """
        x_, y_, z_ = self.cloud()
        x, y, z = self.rotate(x_ * u.R_sun, y_ * u.R_sun, z_ * u.R_sun)

        return SkyCoord(CartesianRepresentation(x, y, z),
                        frame=frames.HeliographicStonyhurst,
                        observer=self.center.observer,
                        obstime=self.center.obstime)

    def rotate(self, x_, y_, z_):
        """
        Rotates the (x,y,z) coordinates from the geometrical model's coordinate system
        to the Heliospheric coordinates.

        Returns the coordinates as `~astropy.coordinates.SkyCoord`.
        """
        Longc = self.center.lon.to_value(u.rad)
        Latc = self.center.lat.to_value(u.rad)
        tilt = self.tilt.to_value(u.rad)

        v = np.transpose([x_.flatten(), y_.flatten(), z_.flatten()])
        v = Rotation.from_euler('zyz', [tilt, -Latc + np.pi/2, Longc]).apply(v)

        x = np.reshape(v[:, 0], x_.shape) * x_.unit
        y = np.reshape(v[:, 1], y_.shape) * y_.unit
        z = np.reshape(v[:, 2], z_.shape) * z_.unit

        return x, y, z

    def shell_skeleton(self):
        """
        Computes the skeleton axis of the GCS CME model.
        Converted and modified from the IDL script shellskeleton.pro
        https://hesperia.gsfc.nasa.gov/ssw/stereo/secchi/idl/scraytrace/shellskeleton.pro
        """
        self.height.to_value(u.R_sun)
        alpha = self.alpha.to_value(u.rad)
        kappa = self.kappa
        h = self.h.to_value(u.R_sun)
        gamma = np.arcsin(self.kappa)

        # calculate the points of the straight line part
        pRstart = np.array([0, np.sin(alpha), np.cos(alpha)])
        pLstart = np.array([0, -np.sin(alpha), np.cos(alpha)])
        vertsl = np.linspace(0, h, self.nbvertsl)
        pslR = np.outer(vertsl, pRstart)
        pslL = np.outer(vertsl, pLstart)
        rsl = np.tan(gamma) * norm(pslR, axis=1)
        casl = np.full(self.nbvertsl, -alpha)

        # calculate the points of the circular part
        beta = np.linspace(-alpha, np.pi/2, self.nbvertcirc)
        hf = h
        h = hf / np.cos(alpha)
        rho = hf * np.tan(alpha)

        X0 = (rho + h * kappa ** 2 * np.sin(beta)) / (1 - kappa ** 2)
        rc = np.sqrt((h ** 2 * kappa ** 2 - rho ** 2) / (1 - kappa ** 2) + X0 ** 2)
        cac = beta

        pcR = np.array([np.zeros(beta.shape), X0 * np.cos(beta), h + X0 * np.sin(beta)]).T
        pcL = np.array([np.zeros(beta.shape), -X0 * np.cos(beta), h + X0 * np.sin(beta)]).T

        # This part is from here
        # https://gitlab.physik.uni-kiel.de/ET/gcs_python/-/blob/master/gcs/geometry.py
        r = np.concatenate((rsl, rc[1:], np.flipud(rc)[1:], np.flipud(rsl)[1:]))
        ca = np.concatenate((casl, cac[1:], np.pi-np.flipud(cac)[1:], np.pi-np.flipud(casl)[1:]))
        p = np.concatenate((pslR, pcR[1:], np.flipud(pcL)[1:], np.flipud(pslL)[1:]))

        return p, r, ca

    def cloud(self):
        """
        Returns the cloud of points for the GCS model.

        Notes
        -----
        Converted and modified from the IDL script cmecloud.pro.
        https://hesperia.gsfc.nasa.gov/ssw/stereo/secchi/idl/scraytrace/cmecloud.pro
        https://gitlab.physik.uni-kiel.de/ET/gcs_python/-/blob/master/gcs/geometry.py
        """

        # Compute the shell's skeleton axis
        p, r, ca = self.shell_skeleton()

        # This part is from here
        # https://gitlab.physik.uni-kiel.de/ET/gcs_python/-/blob/master/gcs/geometry.py
        ptheta = np.linspace(0, 2*np.pi, self.nbvertcircshell)
        pphi = np.arange(0, 2*(self.nbvertcirc + self.nbvertsl) - 3)
        uv, vv = np.meshgrid(ptheta, pphi)
        uv, vv = uv.flatten(), vv.flatten()
        mesh = r[vv, np.newaxis] * np.array([np.cos(uv), np.sin(uv) * np.cos(ca[vv]), np.sin(uv) * np.sin(ca[vv])]).T + p[vv]

        return mesh[:, 0], mesh[:, 1], mesh[:, 2]

    def plot(self, axis, mode=False, only_surface=False):
        '''
        Plots the GCS model at the provided axes
        '''
        axis.plot_coord(self.coordinates, color='green', linestyle='-', linewidth=0.4)

    def to_dataframe(self):
        """
        Returns the GCS parameters as `~pandas.DataFrame`.
        """
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        data_dict = {
            'hgln': self.center.lon.to_value(u.degree),
            'hglt': self.center.lat.to_value(u.degree),
            'crln': center_.lon.to_value(u.degree),
            'crlt': center_.lat.to_value(u.degree),
            'rcenter': self.rcenter.to_value(u.R_sun),
            'height': self.height.to_value(u.R_sun),
            'rapex': self.rapex.to_value(u.R_sun),
            'alpha': self.alpha.to_value(u.degree),
            'kappa': self.kappa,
            'tilt': self.tilt.to_value(u.degree),
        }

        return pd.DataFrame(data_dict, index=[self.center.obstime.to_datetime()])

    def __str__(self):
        '''
        Returns the GCS object and parameters in a printable sting array.
        '''
        center_ = self.center.transform_to(frames.HeliographicCarrington)
        output = '<GCS object \n'
        output += 'HGLN = %3.2f degrees \n'%self.center.lon.to_value(u.degree)
        output += 'HGLT = %3.2f degrees \n'%self.center.lat.to_value(u.degree)
        output += 'CRLN = %3.2f degrees \n'%center_.lon.to_value(u.degree)
        output += 'CRLT = %3.2f degrees \n'%center_.lat.to_value(u.degree)
        output += 'rcenter = %3.2f Rsun \n'%self.rcenter.to_value(u.R_sun)
        output += 'height = %3.2f Rsun \n'%self.radaxis.to_value(u.R_sun)
        output += 'alpha = %3.2f \n'%self.alpha
        output += 'kappa = %3.2f \n'%self.kappa
        output += 'tilt = %3.2f '%self.tilt
        output += '>'
        return output


def my_plot_coord(coord, axis, **kargs):

    ass = coord.spherical.distance > 1*u.R_sun  # above solar surface
    s = ''.join('X' if p else 'O' for p in ass)
    clist = list([m.span()[0], abs(operator.sub(*m.span()))] for m in re.finditer('X+', s))
    for c in clist:
        axis.plot_coord(coord[c[0]:(c[0]+c[1])], **kargs)



def apply(st):
    # Hide the menu button
    st.markdown(""" <style>
                #MainMenu {visibility: hidden;}
                footer {visibility: hidden;}
                </style> """, unsafe_allow_html=True)
    # Do some css styling tricks here (e.g. remove the padding)
    # https://medium.com/ssense-tech/streamlit-tips-tricks-and-hacks-for-data-scientists-d928414e0c16
    padding = 1
    st.markdown(f""" <style>
                .css-12oz5g7{{
                padding-top: {padding}rem;
                margin-top: -3.5rem;
                max-width: 50rem;
                padding-right: {padding}rem;
                padding-left: {padding}rem;
                padding-bottom: {padding}rem;
                }} </style> """, unsafe_allow_html=True)
    st.markdown(f""" <style>
                .css-128j0gw{{
                margin-top: -3.0rem;
                }} </style> """, unsafe_allow_html=True)
    # Reduce the space in horizontal component
    st.markdown(f""" <style>
                hr {{
                margin: 10px 0px;
                }} </style> """, unsafe_allow_html=True)
    # Custom button appearance
    st.markdown('''
                <style>
                div.stButton > button:first-child {
                display: inline-flex;
                align-items: center;
                justify-content: center;
                min-width: 100%;
                background-color: rgb(255, 255, 255);
                color: rgb(38, 39, 48);
                padding: .25rem .75rem;
                position: relative;
                text-decoration: none;
                border-radius: 4px;
                border-width: 1px;
                border-style: solid;
                border-color: rgb(230, 234, 241);
                border-image: initial; }
                div.stButton > button:hover {
                border-color: rgb(246, 51, 102);
                color: rgb(246, 51, 102);}
                div.stButton > button:active{
                box-shadow: none;
                background-color: rgb(246, 51, 102);
                color: white; }
                </style>''', unsafe_allow_html=True)
    # Custom button appearance
    st.markdown('''
                <style>
                div.stDownloadButton > button:first-child {
                display: inline-flex;
                align-items: center;
                justify-content: center;
                min-width: 100%;
                background-color: rgb(255, 255, 255);
                color: rgb(38, 39, 48);
                padding: .25rem .75rem;
                position: relative;
                text-decoration: none;
                border-radius: 4px;
                border-width: 1px;
                border-style: solid;
                border-color: rgb(230, 234, 241);
                border-image: initial;}
                div.stDownloadButton > button:hover {
                border-color: rgb(246, 51, 102);
                color: rgb(246, 51, 102);}
                div.stDownloadButton > button:active {
                box-shadow: none;
                background-color: rgb(246, 51, 102);
                color: white;}
                </style>''', unsafe_allow_html=True)



"""
This submodule provides utility functions to act on `sunpy.map.GenericMap` instances.
"""

import copy
import warnings

import astropy.units as u
import numpy as np
import sunpy.map
import sunpy_dev.extern.sunkit_instruments.stereo.utils

__all__ = ['maps_sequence_processing', 'get_closest', 'normalize_exposure',
           'prepare_maps', 'difference_maps', 'mask_occulter']


def maps_sequence_processing(map_sequence, seq_type='Plain'):
    """
    Returns a sequence of maps which is processed as running difference or base difference or plain images.

    Parameters
    ----------
    map_sequence : `~sunpy.map.GenericMap`
        A list SunPy maps.

    seq_type : '~str'
        The type of sequence processing

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.
    """

    normalized = True if True in [tmap.exposure_time == 1.0*u.second for tmap in map_sequence] else False
    if not normalized:
        warnings.warn('Warning [maps_sequence_processing]: The exposure time of the maps is not normalized.')

    smap = []
    if seq_type == 'Running Diff.':
        for i in range(1, len(map_sequence)):
            smap_diff = difference_maps(map_sequence[i], map_sequence[i-1])
            smap.append(smap_diff)
    if seq_type == 'Base Diff.':
        for i in range(1, len(map_sequence)):
            smap_diff = difference_maps(map_sequence[i], map_sequence[0])
            smap.append(smap_diff)
    if seq_type == 'Plain':
        for i in range(0, len(map_sequence)):
            smap.append(map_sequence[i])

    if len(smap) != 0:
        sequence_final = sunpy.map.Map(smap, sequence=True)
    else:
        sequence_final = []

    return sequence_final


def get_closest(smap, date):
    """
    Returns the map closest to the date provided from a sequence
    or a list of smaps.

    Parameters
    ----------
    smap : `~sunpy.map.GenericMap`
        A SunPy map.

    date : 'astropy.time.Time'
        A time object.

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.
    """
    delta_time = [abs(map_.date-date) for map_ in smap]
    map_closest = smap[delta_time.index(min(delta_time))]
    return map_closest


def normalize_exposure(smap):
    """
    Normalize the exposure time of a map.

    Parameters
    ----------
    smap : `~sunpy.map.GenericMap`
        A SunPy map.

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.
    """
    if smap.exposure_time == 1*u.second:
        return smap
    new_meta = copy.deepcopy(smap.meta)
    if 'exptime' in new_meta:
        new_meta['exptime'] = 1.0
    elif 'xposure' in new_meta:
        new_meta['xposure'] = 1.0
    smap_new = sunpy.map.Map(smap.data/smap.exposure_time.to_value(u.second), new_meta)
    if smap_new.exposure_time != 1*u.second:
        raise Exception('Could not normalize Maps exposure time.')
    return smap_new


def filter_maps(map_sequence, extra):
    '''
    Returns filtered maps.

    For a map (or a sequence of maps) it performs a filtering, namely:
    removes map dublicates, maps with non-default exposure or dimensions, or polarized maps,
    combine the polarization images to total brightness,

    Parameters
    ----------
    map_sequence : `~sunpy.map.GenericMap`
        A SunPy map.

    extra : A list with with the preparation arguments

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.

    '''
    indices = []
    # TODO: This has been added because VSO search returns duplicates for WISPR
    #       so we manualy filter the dublicates until this problem is solved https://github.com/sunpy/sunpy/issues/5481
    if 'dublicates' in extra:
        map_sequence.sort(key=attrgetter('date'))
        for i in range(0, len(map_sequence)-1):
            if map_sequence[i].date == map_sequence[i+1].date:
                indices.append(i)
        for i in sorted(indices, reverse=True):
            map_sequence.pop(i)

    if 'exposure' in extra:
        map_sequence = [tmap for tmap in map_sequence if tmap.exposure_time > extra['exposure']*u.second]

    if 'dimensions' in extra:
        map_sequence = [tmap for tmap in map_sequence if (tmap.dimensions[0], tmap.dimensions[1]) == extra['dimensions']]

    if 'polar' in extra:
        map_sequence = [tmap for tmap in map_sequence if tmap.meta['polar'] == extra['polar']]

    if len(map_sequence) != 0:
        sequence_final = sunpy.map.Map(map_sequence, sequence=True)
    else:
        sequence_final = []
    return sequence_final


def prepare_maps(map_sequence, extra):
    '''
    Returns prepared maps.

    For a map (or a sequence of maps) it performs basic preparations, namely:
    exposure time normalization, mask the occulters, downsample with superpixel,
    combine the polarization images to total brightness,

    Parameters
    ----------
    map_sequence : `~sunpy.map.GenericMap`
        A SunPy map.

    extra : A list with with the preparation arguments

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.

    '''
    if len(map_sequence) == 0:
        return []

    map_sequence = [normalize_exposure(tmap) for tmap in map_sequence]

    map_sequence = [mask_occulter(tmap) for tmap in map_sequence]

    if 'superpixel' in extra:
        nsuper = extra['superpixel']
        super_dim = u.Quantity([nsuper, nsuper] * u.pixel)
        map_sequence = [tmap.superpixel(super_dim) for tmap in map_sequence]

    # map_sequence = [tmap.rotate(recenter=True) for tmap in map_sequence]

    if map_sequence[0].detector == 'COR1':
        if 'polar' not in extra:
            sequence_final = sunpy_dev.extern.sunkit_instruments.stereo.utils.cor_polariz(map_sequence)
    else:
        sequence_final = map_sequence
    return sequence_final


def difference_maps(smapi, smapm):
    """
    Returns a running or base difference map.

    Parameters
    ----------
    smapi : `~sunpy.map.GenericMap`
        A SunPy map.

    smapm : `~sunpy.map.GenericMap`
        A SunPy map.

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.

    Note
    -------
    The two maps have to be coalligned.
    This is something that it is not checked here.
    """
    if smapi.exposure_time != 1*u.second:
        smapi = normalize_exposure(smapi)
    if smapm.exposure_time != 1*u.second:
        smapm = normalize_exposure(smapm)
    smap_difference = smapi.data - smapm.data
    return sunpy.map.Map(smap_difference, smapi.meta)


def mask_occulter(smap, apply_mask=True, mask_value=0):
    """
    For a given map it produces a mask for the coronagraphic occulters
    and returns the map with the mash applied or the mask. If the map
    is not from coronagraphs or it is not registered returns the initial map

    Parameters
    ----------
    smap : `~sunpy.map.GenericMap`
        A SunPy map.

    Returns
    -------
    `~sunpy.map.GenericMap`
        A SunPy map.
    """
    if smap.detector == 'COR1':
        xcen, ycen = smap.meta['crpix1'], smap.meta['crpix2']
        in_fac, out_fac = (100/512), (280/512)
    elif smap.detector == 'COR2':
        xcen, ycen = smap.meta['crpix1'], smap.meta['crpix2']
        in_fac, out_fac = (150/2048), (1030/2048)
    elif smap.detector == 'C2':
        xcen = (smap.meta['crval1'] / smap.meta['cdelt1']) +  smap.meta['crpix1']
        ycen = (smap.meta['crval2'] / smap.meta['cdelt2']) + smap.meta['crpix2']
        in_fac, out_fac = (170/1024), (670/1024)
    elif smap.detector == 'C3':
        xcen = (smap.meta['crval1'] / smap.meta['cdelt1']) + smap.meta['crpix1']
        ycen = (smap.meta['crval2'] / smap.meta['cdelt2']) + smap.meta['crpix2']
        in_fac, out_fac = (60/1024), (560/1024)
    else:
        if apply_mask:
            return smap
        return []

    inner_distance = in_fac * smap.dimensions.x
    outer_distance = out_fac * smap.dimensions.x

    x, y = np.meshgrid(*[np.arange(v.value) for v in smap.dimensions]) * u.pixel
    xprime_sq = (x-xcen * u.pixel) ** 2
    yprime_sq = (y-ycen * u.pixel) ** 2
    mask_inner = np.sqrt(xprime_sq + yprime_sq) < inner_distance
    mask_outer = np.sqrt(xprime_sq + yprime_sq) > outer_distance

    if apply_mask:
        smap.data[mask_inner+mask_outer] = mask_value
        return sunpy.map.Map(smap.data, smap.meta)
    return mask_inner+mask_outer



"""
    PyThea: A software package to reconstruct the 3D structure of CMEs and
    shock waves using multi-viewpoint remote-sensing observations.
    Copyright (C) 2021  Athanasios Kouloumvakos

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import datetime
import json

import astropy.units as u
import numpy as np
import pandas as pd
from callbacks import change_long_lat_sliders
from config.config_sliders import sliders_dict as sd
from geometrical_models import ellipsoid, gcs, spheroid
from utils import get_hek_flare, model_fittings


def date_and_event_selection(st):
    st.sidebar.markdown('## Date and event selection')
    col1, col2 = st.sidebar.columns(2)
    day = col1.date_input('Select a day to process',
                          value=datetime.date.today() - datetime.timedelta(days=1),
                          min_value=datetime.datetime(2010, 1, 1, 0, 0, 0),
                          max_value=datetime.date.today())
    initialisation = col2.select_slider('How to initialise?',
                                        options=('Manual', 'Event', 'File'), value='Event')
    if initialisation == 'Event':
        selectbox_list, flare_list_ = get_hek_flare(day)
        selectbox_list.insert(0, 'Select event')
        event_selected = st.sidebar.selectbox('Solar Events', options=selectbox_list)
        if event_selected != 'Select event' and event_selected != 'No events returned':
            st.session_state.date_process = flare_list_['event_peaktime'][selectbox_list.index(event_selected)-1]
            st.session_state.event_selected = event_selected
            st.experimental_rerun()
        elif event_selected == 'No events returned':
            st.sidebar.warning('Initiate manually')
    elif initialisation == 'Manual':
        with st.sidebar.form('my_form'):
            col1, col2 = st.columns(2)
            ev_id = col1.selectbox('Event ID',
                                   options=['Select', 'FL', 'CME', 'SHW'])
            time = col2.time_input('Time', datetime.time(12, 0))
            if ev_id != 'Select':
                st.session_state.date_process = datetime.datetime.combine(day, time)
                t_ = st.session_state.date_process.strftime('%Y-%m-%dT%H:%M:%S')
                st.session_state.event_selected = f'{ev_id}|{t_}'
                st.experimental_rerun()
            st.form_submit_button()
    elif initialisation == 'File':
        uploaded_file = st.sidebar.file_uploader('Provide a fitting file', accept_multiple_files=False)
        if uploaded_file:
            if uploaded_file.type != 'application/json':
                st.sidebar.error('Unsupported file type')
                st.stop()
            fitting = json.loads(uploaded_file.read())
            st.session_state.event_selected = fitting['event_selected']
            st.session_state.date_process  = datetime.datetime.strptime(fitting['date_process'], '%Y-%m-%dT%H:%M:%S.%f')
            st.session_state.geometrical_model = fitting['geometrical_model']['type']
            table_indx = [datetime.datetime.strptime(t, '%Y-%m-%dT%H:%M:%S.%f') for t in fitting['geometrical_model']['parameters']['time']]
            parameters = pd.DataFrame(fitting['geometrical_model']['parameters'], index=table_indx)
            parameters = parameters.drop(['time'], axis=1)

            if 'kinematics' in fitting:
                kinematics = fitting['kinematics']
            else:
                # TODO: Remove this in version 1.0.0
                st.warning('**Warning:** The .json fitting file does not contain the "kinematics" information. \
                            This means that the file was prodused using Pythea with <V0.6.0. \
                            **To resolve this:** Just do a save of the loaded fitting now and replace \
                            the old file with the new one. This will not alter your fittings.')
                kinematics = {'fit_method': {'type': 'polynomial', 'order': 1}}

            st.session_state.model_fittings = model_fittings(fitting['event_selected'],
                                                             fitting['date_process'],
                                                             fitting['geometrical_model']['type'],
                                                             parameters,
                                                             kinematics=kinematics)
            st.experimental_rerun()


def fitting_and_slider_options_container(st):
    container = st.sidebar.container()

    with container.expander('Options'):
        col1, col2 = st.columns(2)
        col1.radio('Coordinate system', options=['HGS', 'HGC'],
                   on_change=change_long_lat_sliders, args=[st], key='coord_system')

        if st.session_state.geometrical_model == 'Spheroid':
            col2.radio('Axes representation', options=['h, e, k', 'r, a, b'], key='sliders_repr_mode')
        elif st.session_state.geometrical_model == 'Ellipsoid':
            col2.radio('Axes representation', options=['h, e, k, a', 'r, a, b, c'], key='sliders_repr_mode')
        elif st.session_state.geometrical_model == 'GCS':
            col2.radio('Axes representation', options=['h, a, k, t'], key='sliders_repr_mode')
        else:
            st.error('Unrecognized geometrical model')

        st.selectbox('Adjustments mode',
                     options=['Standard', '<10Rsun', '>10Rsun', '>30Rsun'],
                     key='sliders_adjustment_mode')

        st.select_slider('Plot option for fitting mesh',
                         options=['No plot', 'Skeleton', 'Full', 'Surface'],
                         value='Skeleton', key='plot_mesh_mode')

        st.checkbox('View Fitting Table', value=False, key='fitting_table')

        st.checkbox('View Kinematics Plot', value=False, key='plt_kinematics')

        if 'fit_args_prime' not in st.session_state and\
           'model_fittings' in st.session_state:
            if st.session_state.model_fittings.kinematics['fit_method'] is None:
                # TODO: When initialization is 'Event' or 'Manual' the st.session_state.model_fittings.kinematics
                # does not take a default value. Explore if I can move this earlier.
                st.session_state.model_fittings.kinematics = {'fit_method': {'type': 'polynomial', 'order': 1}}
            st.session_state.fit_args_prime = st.session_state.model_fittings.kinematics['fit_method']

        if (('model_fittings' in st.session_state) and
           (st.session_state.plt_kinematics is True) and
           st.button('Load Fitting Parameters', key='load_kinematics_param')) or\
           (st.session_state.startup['fitting'] and
           'fit_args_prime' in st.session_state):
            st.session_state.startup['fitting'] = False
            fit_opt = st.session_state.fit_args_prime
            st.session_state.fit_mode = fit_opt['type']
            if fit_opt['type'] == 'polynomial':
                st.session_state.polyfit_order = fit_opt['order']
            elif fit_opt['type'] == 'spline':
                st.session_state.splinefit_order = fit_opt['order']
                st.session_state.splinefit_smooth = fit_opt['smooth']
            elif fit_opt['type'] == 'custom':
                st.session_state.fitcustexpres_select = fit_opt['expression']
                st.session_state.splinefit_order = fit_opt['order']


def fitting_sliders(st):
    options = {
        'Spheroid': {'h, e, k': ['height', 'kappa', 'epsilon'],
                     'r, a, b': ['rcenter', 'radaxis', 'orthoaxis1']
                     },
        'Ellipsoid': {'h, e, k, a': ['height', 'kappa', 'epsilon', 'alpha', 'tilt'],
                      'r, a, b, c': ['rcenter', 'radaxis', 'orthoaxis1', 'orthoaxis2', 'tilt']
                      },
        'GCS': {'h, a, k, t': ['height', 'alpha', 'kappa', 'tilt'],
                },
    }

    adjustments = st.session_state.sliders_adjustment_mode

    if st.session_state.geometrical_model == 'Spheroid' or \
       st.session_state.geometrical_model == 'Ellipsoid' or \
       st.session_state.geometrical_model == 'GCS':
        gmodel = st.session_state.geometrical_model
        rmode = st.session_state.sliders_repr_mode
        sliders = options[gmodel][rmode]
        for slider in sliders:
            if slider not in st.session_state:
                st.session_state[slider] = sd[gmodel][slider][adjustments]['def']
            st.sidebar.slider(f'{slider} {sd[gmodel][slider]["unit"]}:',  # Add units here
                              min_value=sd[gmodel][slider][adjustments]['min'],
                              max_value=sd[gmodel][slider][adjustments]['max'],
                              step=sd[gmodel][slider][adjustments]['step'], key=slider)  # * u.R_sun


def final_parameters_gmodel(st):
    if st.session_state.geometrical_model == 'Spheroid':
        if st.session_state.sliders_repr_mode == 'h, e, k':
            rcenter, radaxis, orthoaxis1 = spheroid.rab_from_hek(st.session_state.height * u.R_sun,
                                                                 st.session_state.epsilon,
                                                                 st.session_state.kappa)
        elif st.session_state.sliders_repr_mode == 'r, a, b':
            rcenter, radaxis, orthoaxis1 = st.session_state.rcenter * u.R_sun, \
                st.session_state.radaxis * u.R_sun, \
                st.session_state.orthoaxis1 * u.R_sun
        return rcenter, radaxis, orthoaxis1
    elif st.session_state.geometrical_model == 'Ellipsoid':
        if st.session_state.sliders_repr_mode == 'h, e, k, a':
            rcenter, radaxis, orthoaxis1, orthoaxis2 = ellipsoid.rabc_from_heka(st.session_state.height * u.R_sun,
                                                                                st.session_state.epsilon,
                                                                                st.session_state.kappa,
                                                                                st.session_state.alpha)
        elif st.session_state.sliders_repr_mode == 'r, a, b, c':
            rcenter, radaxis, orthoaxis1, orthoaxis2 = st.session_state.rcenter * u.R_sun, \
                st.session_state.radaxis * u.R_sun, \
                st.session_state.orthoaxis1 * u.R_sun, \
                st.session_state.orthoaxis2 * u.R_sun
        tilt = st.session_state.tilt * u.degree
        return rcenter, radaxis, orthoaxis1, orthoaxis2, tilt
    elif st.session_state.geometrical_model == 'GCS':
        if st.session_state.sliders_repr_mode == 'h, a, k, t':
            height = st.session_state.height * u.R_sun
            alpha = st.session_state.alpha * u.degree
            kappa = st.session_state.kappa
            rcenter = gcs.rcenter_(height, alpha, kappa)
        tilt = st.session_state.tilt * u.degree
        return rcenter, height, alpha, kappa, tilt


def maps_clims(st, imagers_list):
    if 'map_clim' not in st.session_state:
        st.session_state['map_clim'] = {}

    for imager in imagers_list:
        if imager not in st.session_state.map or st.session_state.map[imager] == []:
            pass
        else:
            map_ = st.session_state.map[imager][0]
            st.session_state.map_clim[imager] = [np.nanquantile(map_.data, 0.20), np.nanquantile(map_.data, 0.80)]



"""
Test the remote clients
"""

import pytest
from sunpy.net import attrs, hek


@pytest.mark.remote_data
def test_hek_client():
    """
    Tests that the HEK client returns a HEKTable
    """
    startTime = '2011/08/09 07:23:56'
    endTime = '2011/08/09 12:40:29'
    eventType = 'FL'

    hekTime = attrs.Time(startTime, endTime)
    hekEvent = attrs.hek.EventType(eventType)

    h = hek.HEKClient()
    hek_query = h.search(hekTime, hekEvent)
    assert type(hek_query) == hek.hek.HEKTable

    # Test that the hek_query['event_peaktime'] is <class 'astropy.time.core.Time'>



import os

from setuptools import find_packages, setup

# from setuptools.config import read_configuration

# extras = read_configuration('pyproject.toml')

# create home directory
if not os.path.isdir(os.path.join(os.environ['HOME'], '.PyThea')):
    os.mkdir(os.path.join(os.environ['HOME'], '.PyThea'))

with open('README_pypi.md', 'r') as f:
    long_description = f.read()

# get requirements
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='PyThea',
    use_scm_version={'write_to': os.path.join('PyThea', '_version.py')},

    description='PyThea: A software package to reconstruct the 3D structure of CMEs and shock waves',
    url='https://github.com/AthKouloumvakos/PyThea',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Athanasios Kouloumvakos',
    author_email='athkouloumvakos@gmail.com',
    license='GPL-3.0',
    license_file='LICENSE.md',
    keywords=['science', 'solar physics', 'solar', 'sun', 'shock waves'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
    ],

    python_requires='>=3.8, <3.10',
    install_requires=requirements,

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'pythea = PyThea.pythea_cli:main'
        ]
    },
    # extras_require=extras,
)
